# Copyright (c) 2018-2020, Eduardo Rodrigues and Henry Schreiner.
#
# Distributed under the 3-clause BSD license, see accompanying file LICENSE
# or https://github.com/scikit-hep/decaylanguage for details.

import decaylanguage


def test_main():
    assert decaylanguage is not None



# Copyright (c) 2018-2020, Eduardo Rodrigues and Henry Schreiner.
#
# Distributed under the 3-clause BSD license, see accompanying file LICENSE
# or https://github.com/scikit-hep/decaylanguage for details.

from lark import Transformer, Tree

def get_from_parser(parser, key):
    return [v.children for v in parser.find_data(key)]

class AmpGenTransformer(Transformer):
    def constant(self, lines):
        particle, value = lines
        return Tree('constant', [str(particle.children[0]), float(value)])
    def event_type(self, lines):
        return Tree('event_type', [str(p.children[0]) for p in lines])
    def fixed(self, lines):
        return False
    def free(self, lines):
        return True
    def variable(self, lines):
        p, free, value, error = lines
        return Tree('variable', [str(p.children[0]), free, float(value), float(error)])
    def cplx_decay_line(self, lines):
        decay, real, imag = lines
        real_free, real_val, real_err = real.children
        imag_free, imag_val, imag_err = imag.children

        decay['fix'] = not (real_free and imag_free)
        decay['amp'] = complex(float(real_val), float(imag_val))
        decay['err'] = complex(float(real_err), float(imag_err))

        return Tree('cplx_decay_line', decay)

    def decay(self, lines):
        particle, = lines[0].children
        dic = {'name' : str(particle),  'daughters' : []}

        for line in lines[1:]:
            if line.data == 'subdecay':
                dic['daughters'] += line.children
            elif line.data == 'decaytype':
                for children in line.children:
                    if children.data == 'spinfactor':
                        dic['spinfactor'], = children.children
                    elif children.data == 'lineshape':
                        dic['lineshape'], = children.children

        return dic



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

import os


extensions = [
    'sphinx.ext.autodoc',
    'sphinx.ext.autosummary',
    'sphinx.ext.coverage',
    'sphinx.ext.doctest',
    'sphinx.ext.extlinks',
    'sphinx.ext.ifconfig',
    'sphinx.ext.napoleon',
    'sphinx.ext.todo',
    'sphinx.ext.viewcode',
]
if os.getenv('SPELLCHECK'):
    extensions += 'sphinxcontrib.spelling',
    spelling_show_suggestions = True
    spelling_lang = 'en_US'

source_suffix = '.rst'
master_doc = 'index'
project = 'decaylanguage'
year = '2018'
author = 'Henry Fredrick Schreiner III'
copyright = '{0}, {1}'.format(year, author)
version = release = '0.2.0'

pygments_style = 'trac'
templates_path = ['.']
extlinks = {
    'issue': ('https://github.com/scikit-hep/decaylanguage/issues/%s', '#'),
    'pr': ('https://github.com/scikit-hep/decaylanguage/pull/%s', 'PR #'),
}
# on_rtd is whether we are on readthedocs.org
on_rtd = os.environ.get('READTHEDOCS', None) == 'True'

if not on_rtd:  # only set the theme if we're building docs locally
    html_theme = 'sphinx_rtd_theme'

html_use_smartypants = True
html_last_updated_fmt = '%b %d, %Y'
html_split_index = False
html_sidebars = {
   '**': ['searchbox.html', 'globaltoc.html', 'sourcelink.html'],
}
html_short_title = '%s-%s' % (project, version)

napoleon_use_ivar = True
napoleon_use_rtype = False
napoleon_use_param = False



from __future__ import absolute_import

from .particleutils import charge_conjugate_name
from .errors import LineFailure
from .utilities import filter_lines
from .utilities import iter_flatten
from .utilities import split

__all__ = (LineFailure, iter_flatten, split, filter_lines)



# Copyright (c) 2018-2020, Eduardo Rodrigues and Henry Schreiner.
#
# Distributed under the 3-clause BSD license, see accompanying file LICENSE
# or https://github.com/scikit-hep/decaylanguage for details.

import pytest

try:
    from pathlib2 import Path
except ImportError:
    from pathlib import Path

from lark import Tree, Token

from decaylanguage import data

from decaylanguage.dec.dec import DecFileParser
from decaylanguage.dec.dec import DecFileNotParsed, DecayNotFound
from decaylanguage.dec.dec import DecayModelParamValueReplacement
from decaylanguage.dec.dec import ChargeConjugateReplacement
from decaylanguage.dec.dec import get_decay_mother_name
from decaylanguage.dec.dec import get_final_state_particle_names
from decaylanguage.dec.dec import get_model_name
from decaylanguage.dec.dec import get_model_parameters

from decaylanguage.dec.enums import known_decay_models

# New in Python 3
try:
    FileNotFoundError
except NameError:
    FileNotFoundError = IOError


DIR = Path(__file__).parent.resolve()


def test_default_constructor():
    p = DecFileParser()
    assert p is not None

def test_constructor_1_file():
    p = DecFileParser(DIR / '../data/test_example_Dst.dec')

    assert p is not None
    assert len(p._dec_file_names) == 1


def test_constructor_multiple_files():
    p = DecFileParser(DIR / '../data/test_Xicc2XicPiPi.dec',
                      DIR / '../data/test_Bc2BsPi_Bs2KK.dec')
    p.parse()

    assert len(p._dec_file_names) == 2
    assert p.number_of_decays == 7


def test_from_string():
    s = """Decay pi0
0.988228297   gamma   gamma                   PHSP;
0.011738247   e+      e-      gamma           PI0_DALITZ;
0.000033392   e+      e+      e-      e-      PHSP;
0.000000065   e+      e-                      PHSP;
Enddecay
"""

    p = DecFileParser.from_string(s)
    p.parse()

    assert p.number_of_decays == 1


def test_unknown_decfile():
    with pytest.raises(FileNotFoundError):
        p = DecFileParser('non-existent.dec')


def test_non_parsed_decfile():
    with pytest.raises(DecFileNotParsed):
        p = DecFileParser(DIR / '../data/test_example_Dst.dec')
        p.list_decay_mother_names()


def test_non_existent_decay():
    with pytest.raises(DecayNotFound):
        p = DecFileParser(DIR / '../data/test_example_Dst.dec')
        p.parse()
        p.list_decay_modes('XYZ')


def test_default_grammar_loading():
    p = DecFileParser(DIR / '../data/test_example_Dst.dec')
    assert p.grammar is not None


def test_explicit_grammar_loading():
    p = DecFileParser(DIR / '../data/test_example_Dst.dec')
    p.load_grammar(DIR / '../../decaylanguage/data/decfile.lark')

    assert p.grammar_loaded is True


def test_string_representation():
        p = DecFileParser(DIR / '../data/test_example_Dst.dec')

        assert "n_decays" not in p.__str__()

        p.parse()
        assert "n_decays=5" in p.__str__()


def test_definitions_parsing():
    p = DecFileParser(DIR / '../data/defs-aliases-chargeconj.dec')
    p.parse()

    assert len(p.dict_definitions()) == 24


def test_aliases_parsing():
    p = DecFileParser(DIR / '../data/defs-aliases-chargeconj.dec')
    p.parse()

    assert len(p.dict_aliases()) == 132


def test_charge_conjugates_parsing():
    p = DecFileParser(DIR / '../data/defs-aliases-chargeconj.dec')
    p.parse()

    assert len(p.dict_charge_conjugates()) == 77


def test_pythia_definitions_parsing():
    p = DecFileParser(DIR / '../data/defs-aliases-chargeconj.dec')
    p.parse()

    assert p.dict_pythia_definitions() == {'ParticleDecays:mixB': 'off',
                                           'Init:showChangedSettings': 'off',
                                           'Init:showChangedParticleData': 'off',
                                           'Next:numberShowEvent': 0.0}


def test_jetset_definitions_parsing():
    p = DecFileParser(DIR / '../data/defs-aliases-chargeconj.dec')
    p.parse()

    assert p.dict_jetset_definitions() == {'MSTU': {1: 0, 2: 0},
                                           'PARU': {11: 0.001},
                                           'MSTJ': {26: 0},
                                           'PARJ': {21: 0.36}}


def test_list_lineshape_definitions():
    p = DecFileParser(DIR / '../data/defs-aliases-chargeconj.dec')
    p.parse()

    assert p.list_lineshape_definitions() == [(['D_1+', 'D*+', 'pi0'], 2),
                                              (['D_1+', 'D*0', 'pi+'], 2),
                                              (['D_1-', 'D*-', 'pi0'], 2),
                                              (['D_1-', 'anti-D*0', 'pi-'], 2),
                                              (['D_10', 'D*0', 'pi0'], 2),
                                              (['D_10', 'D*+', 'pi-'], 2),
                                              (['anti-D_10', 'anti-D*0', 'pi0'], 2),
                                              (['anti-D_10', 'D*-', 'pi+'], 2)]


def test_global_photos_flag():
    p = DecFileParser(DIR / '../data/defs-aliases-chargeconj.dec')
    p.parse()

    assert p.global_photos_flag() == True

def test_missing_global_photos_flag():
    p = DecFileParser(DIR / '../data/test_example_Dst.dec')
    p.parse()

    assert p.global_photos_flag() == False

def test_list_charge_conjugate_decays():
    p = DecFileParser(DIR / '../data/test_Bd2DmTauNu_Dm23PiPi0_Tau2MuNu.dec')
    p.parse()

    assert p.list_charge_conjugate_decays() == ['MyD+', 'MyTau+', 'Mya_1-', 'anti-B0sig']


def test_simple_dec():
    p = DecFileParser(DIR / '../data/test_example_Dst.dec')
    p.parse()

    assert p.list_decay_mother_names() == ['D*+', 'D*-', 'D0', 'D+', 'pi0']

    assert p.list_decay_modes('D0') == [['K-', 'pi+']]


def test_with_missing_info():
    """
    This decay file misses a ChargeConj statement relating the particle aliases
    Xi_cc+sig and anti-Xi_cc-sig. As a consequence, only 3 decays are parsed
    and the following warning is issued:
    ``
    Corresponding 'Decay' statement for 'CDecay' statement(s) of following particle(s) not found:
    anti-Xi_cc-sig.
    Skipping creation of these charge-conjugate decay trees.
      warnings.warn(msg)
    ``
    """
    p = DecFileParser(DIR / '../data/test_Xicc2XicPiPi.dec')
    p.parse()

    # Decay of anti-Xi_cc-sig missing
    assert p.number_of_decays == 3
    assert 'anti-Xi_cc-sig' not in p.list_decay_mother_names()

    # CDecay statements
    assert 'anti-Xi_cc-sig' in p.list_charge_conjugate_decays()


def test_decay_mode_details():
    p = DecFileParser(DIR / '../data/test_example_Dst.dec')
    p.parse()

    tree_Dp = p._find_decay_modes('D+')[0]
    output = (1.0, ['K-', 'pi+', 'pi+', 'pi0'], 'PHSP', '')
    assert p._decay_mode_details(tree_Dp) == output


def test_decay_model_parsing():
    """
    This module tests building blocks rather than the API,
    hence the "strange" way to access parsed Lark Tree instances.
    """
    p = DecFileParser(DIR / '../data/test_Bd2DstDst.dec')
    p.parse()

    # Simple decay model without model parameters
    dl = p._parsed_decays[2].children[1]  # 'MySecondD*+' Tree
    assert get_model_name(dl) == 'VSS'
    assert get_model_parameters(dl) == ''

    # Decay model with a set of floating-point model parameters
    dl = p._parsed_decays[0].children[1]  # 'B0sig' Tree
    assert get_model_name(dl) == 'SVV_HELAMP'
    assert get_model_parameters(dl) == [0.0, 0.0, 0.0, 0.0, 1.0, 0.0]

    # Decay model where model parameter is a string,
    # which matches an XML file for EvtGen
    dl = p._parsed_decays[4].children[1]  # 'MyD0' Tree
    assert get_model_name(dl) == 'LbAmpGen'
    assert get_model_parameters(dl) == ['DtoKpipipi_v1']


def test_decay_model_parsing_with_variable_defs():
    """
    In this example the decay model details are "VSS_BMIX dm",
    where dm stands for a variable name whose value is defined via the statement
    'Define dm 0.507e12'. The parser should recognise this and return
    [0.507e12] rather than ['dm'] as model parameters.
    """
    p = DecFileParser(DIR / '../data/test_Upsilon4S2B0B0bar.dec')
    p.parse()

    assert p.dict_definitions() == {'dm': 507000000000.0}

    dl = p._parsed_decays[0].children[1]
    assert get_model_name(dl) == 'VSS_BMIX'
    assert get_model_parameters(dl) == [0.507e12]


def test_duplicate_decay_definitions():
    p = DecFileParser(DIR / '../data/duplicate-decays.dec')
    p.parse()

    assert p.number_of_decays == 2

    assert p.list_decay_mother_names() == ['Sigma(1775)0', 'anti-Sigma(1775)0']


def test_list_decay_modes():
    p = DecFileParser(DIR / '../data/test_example_Dst.dec')
    p.parse()

    assert p.list_decay_modes('D*-') == [['anti-D0', 'pi-'], ['D-', 'pi0'], ['D-', 'gamma']]
    assert p.list_decay_modes('D*(2010)-', pdg_name=True) == [['anti-D0', 'pi-'], ['D-', 'pi0'], ['D-', 'gamma']]


def test_list_decay_modes_on_the_fly():
    """
    Unlike in the example above the charge conjugate decay modes are created
    on the fly from the non-CC. decay.
    """
    p = DecFileParser(DIR / '../data/test_Xicc2XicPiPi.dec')
    p.parse()

    # Parsed directly from the dec file
    assert p.list_decay_modes('MyXic+') == [['p+', 'K-', 'pi+']]

    # Decay mode created on-the-fly from the above
    assert p.list_decay_modes('MyantiXic-') == [['anti-p-', 'K+', 'pi-']]


def test_print_decay_modes():
    p = DecFileParser(DIR / '../data/test_example_Dst.dec')
    p.parse()

    with pytest.raises(DecayNotFound):
        p.print_decay_modes('D*(2010)-')

    p.print_decay_modes('D*(2010)-', pdg_name=True)


def test_build_decay_chains():
    p = DecFileParser(DIR / '../data/test_example_Dst.dec')
    p.parse()

    output = {'D+': [{'bf': 1.0, 'fs': ['K-', 'pi+', 'pi+', 'pi0'], 'model': 'PHSP', 'model_params': ''}]}
    assert p.build_decay_chains('D+', stable_particles=['pi0']) == output


def test_Lark_DecayModelParamValueReplacement_Visitor_no_params():
    t = Tree('decay', [Tree('particle', [Token('LABEL', 'D0')]),
            Tree('decayline', [Tree('value', [Token('SIGNED_NUMBER', '1.0')]),
            Tree('particle', [Token('LABEL', 'K-')]),
            Tree('particle', [Token('LABEL', 'pi+')]),
            Tree('model', [Token('MODEL_NAME', 'PHSP')])])])

    DecayModelParamValueReplacement().visit(t)

    # The visitor should do nothing in this case
    tree_decayline = list(t.find_data('decayline'))[0]
    assert get_model_name(tree_decayline) == 'PHSP'
    assert get_model_parameters(tree_decayline) == ''


def test_Lark_DecayModelParamValueReplacement_Visitor_single_value():
    t = Tree('decay', [Tree('particle', [Token('LABEL', 'Upsilon(4S)')]),
            Tree('decayline', [Tree('value', [Token('SIGNED_NUMBER', '1.0')]),
            Tree('particle', [Token('LABEL', 'B0')]),
            Tree('particle', [Token('LABEL', 'anti-B0')]),
            Tree('model', [Token('MODEL_NAME', 'VSS_BMIX'),
            Tree('model_options', [Token('LABEL', 'dm')])])])])

    DecayModelParamValueReplacement().visit(t)

    # Nothing done since model parameter name has no corresponding
    # 'Define' statement from which the actual value can be inferred
    tree_decayline = list(t.find_data('decayline'))[0]
    assert get_model_name(tree_decayline) == 'VSS_BMIX'
    assert get_model_parameters(tree_decayline) == ['dm']

    dict_define_defs = {'dm': 0.507e12}

    DecayModelParamValueReplacement(define_defs=dict_define_defs).visit(t)

    # The model parameter 'dm' should now be replaced by its value
    assert get_model_name(tree_decayline) == 'VSS_BMIX'
    assert get_model_parameters(tree_decayline) == [507000000000.0]


def test_Lark_DecayModelParamValueReplacement_Visitor_list():
    t = Tree('decay', [Tree('particle', [Token('LABEL', 'B0sig')]),
            Tree('decayline', [Tree('value', [Token('SIGNED_NUMBER', '1.000')]),
            Tree('particle', [Token('LABEL', 'MyFirstD*-')]),
            Tree('particle', [Token('LABEL', 'MySecondD*+')]),
            Tree('model', [Token('MODEL_NAME', 'SVV_HELAMP'),
            Tree('model_options',
                [Tree('value', [Token('SIGNED_NUMBER', '0.0')]),
                Tree('value', [Token('SIGNED_NUMBER', '0.0')]),
                Tree('value', [Token('SIGNED_NUMBER', '0.0')]),
                Tree('value', [Token('SIGNED_NUMBER', '0.0')]),
                Tree('value', [Token('SIGNED_NUMBER', '1.0')]),
                Tree('value', [Token('SIGNED_NUMBER', '0.0')])])])])])

    DecayModelParamValueReplacement().visit(t)

    # The visitor should do nothing in this case
    tree_decayline = list(t.find_data('decayline'))[0]
    assert get_model_name(tree_decayline) == 'SVV_HELAMP'
    assert get_model_parameters(tree_decayline) == [0.0, 0.0, 0.0, 0.0, 1.0, 0.0]


def test_Lark_ChargeConjugateReplacement_Visitor():
    """
    A simple example usage of the ChargeConjugateReplacement implementation
    of a Lark's Visitor, here replacing all particles in a 'decay' Tree
    by their antiparticles.
    """
    t = Tree('decay', [Tree('particle', [Token('LABEL', 'D0')]),
            Tree('decayline', [Tree('value', [Token('SIGNED_NUMBER', '1.0')]),
            Tree('particle', [Token('LABEL', 'K-')]),
            Tree('particle', [Token('LABEL', 'pi+')]),
            Tree('model', [Token('MODEL_NAME', 'PHSP')])])])

    ChargeConjugateReplacement().visit(t)

    assert get_decay_mother_name(t) == 'anti-D0'
    assert get_final_state_particle_names(t.children[1]) == ['K+', 'pi-']


def test_Lark_ChargeConjugateReplacement_Visitor_with_aliases():
    """
    Example with a D0 decay specified via an alias (MyD0).
    As such, it is necessary to state what the particle-antiparticle match is,
    which in decay files would mean the following lines:
       Alias       MyD0        D0
       Alias       MyAnti-D0   anti-D0
       ChargeConj  MyD0        MyAnti-D0
    A dictionary of matches should be passed to the Lark Visitor instance.
    """
    t = Tree('decay', [Tree('particle', [Token('LABEL', 'MyD0')]),
            Tree('decayline', [Tree('value', [Token('SIGNED_NUMBER', '1.0')]),
            Tree('particle', [Token('LABEL', 'K-')]),
            Tree('particle', [Token('LABEL', 'pi+')]),
            Tree('model', [Token('MODEL_NAME', 'PHSP')])])])

    dict_ChargeConj_defs = {'MyD0': 'MyAnti-D0'}

    ChargeConjugateReplacement(charge_conj_defs=dict_ChargeConj_defs).visit(t)

    assert get_decay_mother_name(t) == 'MyAnti-D0'
    assert get_final_state_particle_names(t.children[1]) == ['K+', 'pi-']


def test_creation_charge_conjugate_decays_in_decfile_with_aliases():
    """
    Decay file contains 5 particle decays defined via a 'Decay' statement
    and the 5 charge-conjugate decays defined via a 'CDecay' statement.
    The decay modes for the latter 5 should be created on the fly,
    hence providing in total 10 sets of particle decays parsed.
    """
    p = DecFileParser(DIR / '../data/test_Bd2DstDst.dec')
    p.parse()

    assert p.number_of_decays == 10


def test_creation_charge_conjugate_decays_in_decfile_without_CDecay_defs():
    """
    Decay file contains 5 particle decays defined via a 'Decay' statement,
    but no related charge-conjugate (CC) decays defined via a 'CDecay' statement
    since the CC particle decays are also defined via a 'Decay' statement
    for all cases except self-conjugate (mother) particles, obviously ;-).
    This being said, this decay file is in fact incomplete by itself,
    as there are no instructions on how to decay the anti-D0 and the D-!
    In short, there should only be 5 sets of decay modes parsed.
    """
    p = DecFileParser(DIR / '../data/test_example_Dst.dec')
    p.parse()

    assert p.number_of_decays == 5


def test_master_DECAYdotDEC_file():
    p = DecFileParser(DIR / '../../decaylanguage/data/DECAY_LHCB.DEC')
    p.parse()

    assert p.number_of_decays == 506

def test_BELLE2_decfile():
    p = DecFileParser(DIR / '../../decaylanguage/data/DECAY_BELLE2.DEC')
    p.parse()

    # Just check the dec file will parse since I do not know
    # how many decays are in the dec file.
    assert p.number_of_decays == 356

def test_lark_file_model_list_consistency():
    """
    Make sure that the list of known decay models in the grammar file
    'decaylanguage/data/decfile.lark' is consistent with that provided
    to the user via
    'from decaylanguage.dec.enums import known_decay_models'.
    """
    filename = str(DIR / '../../decaylanguage/data/decfile.lark')
    with open(filename) as lark_file:
        lines = lark_file.readlines()
        for line in lines:
            if 'MODEL_NAME.2' in line: break
        models = line.split(':')[1].strip(' ').strip('\n').split('"|"')
        models = [m.strip('"') for m in models]

        assert models == list(known_decay_models)



# Copyright (c) 2018-2020, Eduardo Rodrigues and Henry Schreiner.
#
# Distributed under the 3-clause BSD license, see accompanying file LICENSE
# or https://github.com/scikit-hep/decaylanguage for details.

"""
Submodule with classes and utilities to visualize decay chains.
Decay chains are typically provided by the parser of .dec decay files,
see the `DecFileParser` class.
"""

import itertools
try:
    counter = itertools.count().__next__
except AttributeError:
    counter = itertools.count().next

try:
    import pydot
except ImportError:
    raise ImportError("You need pydot for this submodule. Please install pydot with for example 'pip install pydot'\n")

from particle import latex_to_html_name
from particle.converters.bimap import DirectionalMaps


_EvtGen2LatexNameMap, _Latex2EvtGenNameMap = DirectionalMaps("EvtGenName", "LaTexName")


class GraphNotBuiltError(RuntimeError):
    pass


class DecayChainViewer(object):
    """
    The class to visualize a decay chain.

    Example
    -------
    >>> dfp = DecFileParser('my-Dst-decay-file.dec')
    >>> dfp.parse()
    >>> chain = dfp.build_decay_chains('D*+')
    >>> dcv = DecayChainViewer(chain)
    >>> dcv  # display the SVG figure in a notebook
    """

    __slots__ = ("_chain",
                 "_graph",
                 "_graph_attributes")

    def __init__(self, decaychain, **attrs):
        """
        Default constructor.

        Parameters
        ----------
        decaychain: dict
            Input .dec decay file name.
        attrs: optional
            User input pydot.Dot class attributes.

        See also
        --------
        decaylanguage.dec.dec.DecFileParser: class for creating an input decay chain.
        """
        # Store the input decay chain
        self._chain = decaychain

        # Instantiate digraph with defaults possibly overriden by user attributes
        self._graph = self._instantiate_graph(**attrs)

        # Build the actual graph from the input decay chain structure
        self._build_decay_graph()

    def _build_decay_graph(self):
        """
        Recursively navigate the decay chain tree and produce a Graph
        in the DOT language.
        """
        def safe_html_name(name):
            """
            Get a safe HTML name from the EvtGen name.

            Note
            ----
            The match is done using a conversion map rather than via
            `Particle.from_evtgen_name(name).html_name` for 2 reasons:
            - Some decay-file-specific "particle" names (e.g. cs_0)
              are not in the PDG table.
            - No need to load all particle information if all that's needed
              is a match EvtGen - HTML name.
            """
            try:
                return latex_to_html_name(_EvtGen2LatexNameMap[name])
            except:
                return name

        def html_table_label(names, add_tags=False, bgcolor='#9abad6'):
            if add_tags:
                label = '<<TABLE BORDER="0" CELLSPACING="0" BGCOLOR="{bgcolor}">'.format(bgcolor=bgcolor)
            else:
                label = '<<TABLE BORDER="0" CELLSPACING="0" CELLPADDING="0" BGCOLOR="{bgcolor}"><TR>'.format(bgcolor=bgcolor)
            for i, n in enumerate(names):
                if add_tags:
                    label += '<TR><TD BORDER="1" CELLPADDING="5" PORT="p{tag}">{name}</TD></TR>'.format(tag=i, name=safe_html_name(n))
                else:
                    label += '<TD BORDER="0" CELLPADDING="2">{name}</TD>'.format(name=safe_html_name(n))
            label += "{tr}</TABLE>>".format(tr='' if add_tags else '</TR>')
            return label

        def new_node_no_subchain(list_parts):
            label = html_table_label(list_parts, bgcolor='#eef3f8')
            r = 'dec%s' % counter()
            self._graph.add_node(pydot.Node(r, label=label, style='filled', fillcolor='#eef3f8'))
            return r

        def new_node_with_subchain(list_parts):
            list_parts = [list(p.keys())[0] if isinstance(p,dict) else p for p in list_parts]
            label = html_table_label(list_parts, add_tags=True)
            r = 'dec%s' % counter()
            self._graph.add_node(pydot.Node(r, shape='none', label=label))
            return r

        def iterate_chain(subchain, top_node=None, link_pos=None):
            if not top_node: top_node = node_mother
            n_decaymodes = len(subchain)
            for idm in range(n_decaymodes):
                _list_parts = subchain[idm]['fs']
                if not has_subdecay(_list_parts):
                    _ref = new_node_no_subchain(_list_parts)
                    _bf = subchain[idm]['bf']
                    if link_pos is None:
                        self._graph.add_edge(pydot.Edge(top_node, _ref, label=str(_bf)))
                    else:
                        self._graph.add_edge(pydot.Edge('%s:p%s'%(top_node, link_pos), _ref, label=str(_bf)))
                else:
                    _ref_1 = new_node_with_subchain(_list_parts)
                    _bf_1 = subchain[idm]['bf']
                    if link_pos is None:
                        self._graph.add_edge(pydot.Edge(top_node, _ref_1, label=str(_bf_1)))
                    else:
                        self._graph.add_edge(pydot.Edge('%s:p%s'%(top_node, link_pos), _ref_1, label=str(_bf_1)))
                    for i, _p in enumerate(_list_parts):
                        if not isinstance(_p,str):
                            _k = list(_p.keys())[0]
                            iterate_chain(_p[_k], top_node=_ref_1, link_pos=i)

        has_subdecay = lambda ds: not all([isinstance(p,str) for p in ds])

        k = list(self._chain.keys())[0]
        label = html_table_label([k], add_tags=True, bgcolor='#568dba')
        node_mother = pydot.Node("mother",  shape='none', label=label)
        self._graph.add_node(node_mother)
        sc = self._chain[k]

        # Actually build the whole decay chain, iteratively
        iterate_chain(sc)

    def visualize_decay_graph(self, format='png'):
        """
        Visualize the Graph produced, opening the file ('png' by default)
        with the machine default program.
        """
        import tempfile
        import webbrowser
        tmpf = tempfile.NamedTemporaryFile(prefix='DecayChainViewer',
                                           suffix='.{0}'.format(format),
                                           delete=False)
        self.graph.write(tmpf.name, format=format)
        tmpf.close()
        return webbrowser.open(tmpf.name)

    @property
    def graph(self):
        """
        Get the actual Graph. The user now has full control ...
        """
        return self._graph

    def to_string(self):
        """
        Return a string representation of the built Graph in the DOT language.
        The function is a trivial shortcut for `Graph.to_string()`.
        """
        return self.graph.to_string()

    def _instantiate_graph(self, **attrs):
        """
        Return a pydot.Dot class instance using the default attributes
        specified in this class:
        - Default graph attributes are overriden by input by the user.
        - Class and node and edge defaults.
        """
        # Make sure the user cannot override the graph type
        try:
            del attrs['graph_type']
        except KeyError:
            pass

        default_args = self._get_graph_defaults()
        default_args.update(**attrs)

        g = pydot.Dot(**default_args)

        g.set_node_defaults(**self._get_node_defaults())
        g.set_edge_defaults(**self._get_edge_defaults())

        return g

    def _get_graph_defaults(self):
        """
        Note: the graph type cannot be overriden.
        """
        return dict(graph_type='digraph', graph_name='DecayChainGraph',
                    rankdir='LR')

    def _get_node_defaults(self):
        return dict(fontname='Helvetica', fontsize=11, shape='oval')

    def _get_edge_defaults(self):
        return dict(fontcolor='#4c4c4c', fontsize=11)

    def _repr_svg_(self):
        """
        IPython display in SVG format.
        """

        return self.graph.create_svg().decode('UTF-8')



#!/usr/bin/env python
# coding: utf-8
# Copyright (c) 2018-2020, Eduardo Rodrigues and Henry Schreiner.
#
# Distributed under the 3-clause BSD license, see accompanying file LICENSE
# or https://github.com/scikit-hep/decaylanguage for details.

'''
This is a function that takes a filename and either prints out or returns
a string output with the converted set of decay chains and variables.
'''

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import datetime
from functools import partial

from plumbum import colors
from six import StringIO

from particle import SpinType

from decaylanguage.modeling.goofit import GooFitChain
from decaylanguage.modeling.goofit import SF_4Body


def ampgen2goofit(filename, ret_output=False):
    if ret_output:
        output = StringIO()
        printer = partial(print, file=output)
    else:
        printer = print

    lines, all_states = GooFitChain.read_ampgen(str(filename))

    printer(r'/* Autogenerated file by AmpGen2GooFit')
    printer('Generated on ', datetime.datetime.now())

    printer('\n')
    for seen_factor in {p.spindetails() for p in lines}:
        my_lines = [p for p in lines if p.spindetails() == seen_factor]
        printer(colors.bold | seen_factor, ":", *my_lines[0].spinfactors)
        for line in my_lines:
            printer(' ', colors.blue | str(line))

    printer('\n')
    for spintype in SpinType:
        ps = [format(str(p), '11')
              for p in sorted(GooFitChain.all_particles) if p.spin_type == spintype]
        printer("{spintype.name:>12}:".format(spintype=spintype), *ps)

    printer('\n')
    for n, line in enumerate(lines):
        printer('{n:2} {line!s:<70} spinfactors: {lensf}  L: {line.L} [{Lr[0]}-{Lr[1]}]'
                .format(n=n, line=line, lensf=len(line.spinfactors), Lr=line.L_range()))

    # We can make the GooFit Intro code:

    printer(colors.bold & colors.green | '\n\nAll discovered spin configurations:')

    for line in sorted({line.spindetails() for line in lines}):
        printer(colors.green | line)

    printer(colors.bold & colors.blue | '\n\nAll known spin configurations:')

    # TODO: 4 body only
    for e in SF_4Body:
        printer(colors.blue | e.name)

    printer('\n*/\n\n    // Intro')
    printer(GooFitChain.make_intro(all_states))

    printer('\n\n    // Parameters')
    printer(GooFitChain.make_pars())

    # And the lines can be turned into code, as well:

    printer('\n\n    // Lines')
    for n, line in enumerate(lines):
        printer('    // Line', n)
        printer(line.to_goofit(all_states[1:]), end='\n\n\n')

    if(ret_output):
        return output.getvalue()



# Copyright (c) 2018-2020, Eduardo Rodrigues and Henry Schreiner.
#
# Distributed under the 3-clause BSD license, see accompanying file LICENSE
# or https://github.com/scikit-hep/decaylanguage for details.


from __future__ import absolute_import

try:
    from pathlib2 import Path
except ImportError:
    from pathlib import Path

import pytest

from decaylanguage.dec.dec import DecFileParser
from decaylanguage.decay.viewer import DecayChainViewer


DIR = Path(__file__).parent.resolve()


def test_single_decay():
    p = DecFileParser(DIR / '../data/test_example_Dst.dec')
    p.parse()

    chain = p.build_decay_chains('D*+', stable_particles=['D+', 'D0', 'pi0'])
    dcv = DecayChainViewer(chain)
    graph_output_as_dot = dcv.to_string()

    assert 'mother -> dec0  [label="0.677"]' in graph_output_as_dot
    assert 'mother -> dec1  [label="0.307"]' in graph_output_as_dot
    assert 'mother -> dec2  [label="0.016"]' in graph_output_as_dot


def test_simple_decay_chain():
    p = DecFileParser(DIR / '../data/test_example_Dst.dec')
    p.parse()

    chain = p.build_decay_chains('D*+')
    dcv = DecayChainViewer(chain)
    graph_output_as_dot = dcv.to_string()

    assert 'mother -> dec3  [label="0.677"]' in graph_output_as_dot
    assert 'dec3:p0 -> dec4  [label="1.0"]' in graph_output_as_dot
    assert 'mother -> dec5  [label="0.307"]' in graph_output_as_dot
    assert 'dec5:p0 -> dec6  [label="1.0"]' in graph_output_as_dot
    assert 'dec6:p3 -> dec7  [label="0.988228297"]' in graph_output_as_dot


checklist_decfiles = (
    (DIR / '../data/test_Bc2BsPi_Bs2KK.dec', 'B_c+sig'),
    (DIR / '../data/test_Bd2DDst_Ds2DmPi0.dec', 'B0sig'),
    (DIR / '../data/test_Bd2DmTauNu_Dm23PiPi0_Tau2MuNu.dec', 'B0sig'),
    (DIR / '../data/test_Bd2DMuNu.dec', 'anti-B0sig'),
    (DIR / '../data/test_Bd2DstDst.dec', 'anti-B0sig'),
    (DIR / '../data/test_example_Dst.dec', 'D*+'),
    (DIR / '../data/test_Xicc2XicPiPi.dec', 'Xi_cc+sig')
)


@pytest.mark.parametrize("decfilepath,signal_mother", checklist_decfiles)
def test_duplicate_arrows(decfilepath, signal_mother):
    """
    This test effectively checks whether any box node (node with subdecays)
    gets more than one arrow to it, which would show a bug
    in the creation of the DOT file recursively parsing the built decay chain.
    """
    p = DecFileParser(decfilepath, DIR / '../../decaylanguage/data/DECAY_LHCB.DEC')
    p.parse()

    chain = p.build_decay_chains(signal_mother)
    dcv = DecayChainViewer(chain)
    graph_output_as_dot = dcv.to_string()

    l = [i.split(' ')[0] for i in graph_output_as_dot.split('-> dec')[1:]]  # list of node identifiers
    assert len(set(l)) == len(l)


def test_init_non_defaults():
    p = DecFileParser(DIR / '../data/test_example_Dst.dec')
    p.parse()

    chain = p.build_decay_chains('D*+')
    dcv = DecayChainViewer(chain, graph_name='TEST', rankdir='TB')

    assert dcv.graph.get_name() == 'TEST'
    assert dcv.graph.get_rankdir() == 'TB'


def test_graphs_with_EvtGen_specific_names():
    p = DecFileParser(DIR / '../../decaylanguage/data/DECAY_LHCB.DEC')
    p.parse()

    # Not setting many of the particles as stable would result in a gargantuesque chain,
    # which would also take a fair amount of time to build!
    list_stable_particles = ['Xi_c0',
                             'Xi-',
                             'D0',
                             'Omega_c0',
                             'Sigma_c0',
                             'tau-',
                             'D_s-',
                             'J/psi',
                             'pi0',
                             'Lambda0',
                             'psi(2S)'
                             ]

    chain = p.build_decay_chains('Xi_b-', stable_particles=list_stable_particles)
    dcv = DecayChainViewer(chain)

    assert '(cs)<SUB>0</SUB>' in dcv.to_string()  # not 'cs_0' ;-)
    assert '&Xi;<SUB>c</SUB><SUP>0</SUP>' in dcv.to_string()



# Copyright (c) 2018-2020, Eduardo Rodrigues and Henry Schreiner.
#
# Distributed under the 3-clause BSD license, see accompanying file LICENSE
# or https://github.com/scikit-hep/decaylanguage for details.

from particle import Particle, ParticleNotFound

from decaylanguage import data
from decaylanguage.dec import dec
from lark import Lark, Transformer, Tree

import pytest

class TreeToDec2(Transformer):
    missing = set()
    keyerrs = set()

    def __init__(self, alias_dict):
        self.alias_dict = alias_dict

    def particle(self, items):
        label, = items
        if label in self.alias_dict:
            label = self.alias_dict[label]
        try:
            return Particle.from_string(str(label))
        except ParticleNotFound:
            self.missing.add(str(label))
            return str(label)
        except KeyError:
            self.keyerrs.add(str(label))
            return str(label)

@pytest.mark.skip
def test_dec_full():
    with data.open_text(data, 'DECAY_LHCB.DEC') as f:
        txt = f.read()

    with data.open_text(data, 'decfile.lark') as f:
        grammar = f.read()

    l = Lark(grammar, parser='lalr', lexer='standard')  # , transformer = TreeToDec())

    parsed = l.parse(txt)
    assert bool(parsed)

    transformed = dec.TreeToDec().transform(parsed)

    define = dec.define(transformed)
    pythia_def = dec.pythia_def(transformed)
    alias = dec.alias(transformed)
    chargeconj = dec.chargeconj(transformed)
    global_photos = dec.global_photos(transformed)
    decay = dec.decay(transformed)
    cdecay = dec.cdecay(transformed)
    setlspw = dec.setlspw(transformed)

    for item in pythia_def:
        print(item[0], ":", item[1], "=", item[2])


    labelled = TreeToDec2(alias).transform(decay)

    print(TreeToDec2.missing)
    print(TreeToDec2.keyerrs)

    assert len(TreeToDec2.missing) == 0



# Copyright (c) 2018-2020, Eduardo Rodrigues and Henry Schreiner.
#
# Distributed under the 3-clause BSD license, see accompanying file LICENSE
# or https://github.com/scikit-hep/decaylanguage for details.

"""
Submodule with classes and utilities to deal with and parse .dec decay files.

Basic assumptions
-----------------

1) For standard (non-alias) particle (names):
    - Decay modes defined via a 'Decay' statement.
    - Related antiparticle decay modes either defined via a 'CDecay' statement
      or via a 'Decay' statement. The latter option is often used if CP matters.
2) For particle names defined via aliases:
    - Particle decay modes defined as above.
    - Related antiparticle decay modes defined with either options above,
      *but* there needs to be a 'ChargeConj' statement specifying the
      particle-antiparticle match. Typically::

        Alias MyP+ P+
        Alias MyP- P-
        ChargeConj MyP+ MyP-
        Decay MyP+
        ...
        Enddecay
        CDecay MyP-

3) As a consequence, particles that are self-conjugate should not be used
   in 'CDecay' statements, obviously.
"""

from __future__ import absolute_import, division, print_function

import os
import warnings
import re
import operator

from six import StringIO

from lark import Lark
from lark import Tree, Transformer, Visitor

from particle import Particle, ParticleNotFound
from particle.converters import PDG2EvtGenNameMap
from particle.particle.enums import Charge, Charge_undo, Charge_mapping

from .enums import PhotosEnum

from ..utils import charge_conjugate_name
from ..data import open_text
from .. import data


# New in Python 3
try:
    FileNotFoundError
except NameError:
    FileNotFoundError = IOError


class DecFileNotParsed(RuntimeError):
    pass


class DecayNotFound(RuntimeError):
    pass


class DecFileParser(object):
    """
    The class to parse a .dec decay file.

    Example
    -------
    >>> parsed_file = DecFileParser('my-decay-file.dec')
    """

    __slots__ = ("_grammar",
                 "_grammar_info",
                 "_dec_file_names",
                 "_dec_file",
                 "_parsed_dec_file",
                 "_parsed_decays",
                 "_include_ccdecays")

    def __init__(self, *filenames):
        """
        Default constructor. Parse one or more .dec decay files.

        Parameters
        ----------
        filenames: non-keyworded variable length argument
            Input .dec decay file name(s).
        """
        self._grammar = None       # Loaded Lark grammar definition file
        self._grammar_info = None  # Name of Lark grammar definition file

        # Name(s) of the input decay file(s)
        if filenames:
            # Conversion to handle pathlib on Python < 3.6
            self._dec_file_names = [str(f) for f in filenames]

            stream = StringIO()
            for filename in self._dec_file_names:
                # Check input file
                if not os.path.exists(filename):
                    raise FileNotFoundError("'{0}'!".format(filename))

                with open(filename, 'r') as file:
                    for line in file:
                        # We need to strip the unicode byte ordering if present before checking for *
                        beg = line.lstrip('\ufeff').lstrip()
                        # Make sure one discards all lines "End"
                        # in intermediate files, to avoid a parsing error
                        if not ( beg.startswith('End') and not beg.startswith('Enddecay')):
                            stream.write(line)
                    stream.write('\n')

            stream.seek(0)
            # Content of input file(s)
            self._dec_file = stream.read()
        else:
            self._dec_file_names = None
            self._dec_file = None

        self._parsed_dec_file = None  # Parsed decay file
        self._parsed_decays = None    # Particle decays found in the decay file

        # By default, consider charge-conjugate decays when parsing
        self._include_ccdecays = True

    @classmethod
    def from_string(cls, filecontent):
        """
        Parse a .dec decay file provided as a multi-line string.

        Parameters
        ----------
        filecontent: str
            Input .dec decay file content.
        """
        stream = StringIO(filecontent)
        stream.seek(0)

        _cls = cls()
        _cls._dec_file_names = '<dec file input as a string>'
        _cls._dec_file = stream.read()

        return _cls

    def parse(self, include_ccdecays=True):
        """
        Parse the given .dec decay file according to default Lark parser
        and specified options, i.e.,
            parser = 'lalr',
            lexer = 'standard'.

        See method 'load_grammar' for how to explicitly define the grammar
        and set the Lark parsing options.

        Parameters
        ----------
        include_ccdecays: boolean, optional, default=True
            Choose whether or not to consider charge-conjugate decays,
            which are specified via "CDecay <MOTHER>".
            Make sure you understand the consequences of ignoring
            charge conjugate decays - you won't have a complete picture!
        """
        # Has a file been parsed already?
        if self._parsed_decays is not None:
            warnings.warn("Input file being re-parsed ...")

        # Override the parsing settings for charge conjugate decays
        self._include_ccdecays = include_ccdecays if include_ccdecays else False

        # Retrieve all info on the default Lark grammar and its default options,
        # effectively loading it
        opts = self.grammar_info()
        extraopts = dict((k, v) for k, v in opts.items()
                         if k not in ('lark_file', 'parser','lexer'))

        # Instantiate the Lark parser according to chosen settings
        parser = Lark(self.grammar(), parser=opts['parser'], lexer=opts['lexer'])

        self._parsed_dec_file = parser.parse(self._dec_file)

        # At last, find all particle decays defined in the .dec decay file ...
        self._find_parsed_decays()

        # Check whether certain decay model parameters are defined via
        # variable names with actual values provided via 'Define' statements,
        # and perform the replacements name -> value where relevant.
        # Do also a replacement of 'a_float' with a_float.
        dict_define_defs = self.dict_definitions()
        for tree in self._parsed_decays:
            DecayModelParamValueReplacement(define_defs=dict_define_defs).visit(tree)

        # ... and create on the fly the charge conjugate decays, if requested
        if self._include_ccdecays:
            self._add_charge_conjugate_decays()

    def grammar(self):
        """
        Access the internal Lark grammar definition file,
        loading it from the default location if needed.

        Returns
        -------
        out: str
            The Lark grammar definition file.
        """
        if not self.grammar_loaded:
            self.load_grammar()

        return self._grammar

    def grammar_info(self):
        """
        Access the internal Lark grammar definition file name and
        parser options, loading the grammar from the default location if needed.

        Returns
        -------
        out: dict
            The Lark grammar definition file name and parser options.
        """
        if not self.grammar_loaded:
            self.load_grammar()

        return self._grammar_info

    def load_grammar(self, filename=None, parser='lalr', lexer='standard', **options):
        """
        Load a Lark grammar definition file, either the default one,
        or a user-specified one, optionally setting Lark parsing options.

        Parameters
        ----------
        filename: str, optional, default=None
            Input .dec decay file name. By default 'data/decfile.lark' is loaded.
        parser: str, optional, default='lalr'
            The Lark parser engine name.
        lexer: str, optional, default='standard'
            The Lark parser lexer mode to use.
        options: keyword arguments, optional
            Extra options to pass on to the parsing algorithm.

        See Lark's Lark class for a description of available options
        for parser, lexer and options.
        """

        if filename is None:
            filename = 'decfile.lark'
            with data.open_text(data, filename) as f:
                self._grammar = f.read()
        else:
            # Conversion to handle pathlib on Python < 3.6:
            filename = str(filename)

            self._grammar = open(filename).read()

        self._grammar_info = dict(lark_file=filename, parser=parser, lexer=lexer, **options)

    @property
    def grammar_loaded(self):
        """Check to see if the Lark grammar definition file is loaded."""
        return self._grammar is not None

    def dict_definitions(self):
        """
        Return a dictionary of all definitions in the input parsed file,
        of the form "Define <NAME> <VALUE>",
        as {'NAME1': VALUE1, 'NAME2': VALUE2, ...}.
        """
        return get_definitions(self._parsed_dec_file)

    def dict_aliases(self):
        """
        Return a dictionary of all alias definitions in the input parsed file,
        of the form "Alias <NAME> <ALIAS>",
        as {'NAME1': ALIAS1, 'NAME2': ALIAS2, ...}.
        """
        return get_aliases(self._parsed_dec_file)

    def dict_charge_conjugates(self):
        """
        Return a dictionary of all charge conjugate definitions
        in the input parsed file, of the form
        "ChargeConj <PARTICLE> <CC_PARTICLE>", as
        {'PARTICLE1': CC_PARTICLE1, 'PARTICLE2': CC_PARTICLE2, ...}.
        """
        return get_charge_conjugate_defs(self._parsed_dec_file)

    def dict_pythia_definitions(self):
        """
        Return a dictionary of all Pythia definitions in the input parsed file,
        of the form
        "PythiaBothParam <NAME>=<LABEL>"
        or
        "PythiaBothParam <NAME>=<NUMBER>",
        as {'NAME1': 'LABEL1', 'NAME2': VALUE2, ...}.
        """
        return get_pythia_definitions(self._parsed_dec_file)

    def dict_jetset_definitions(self):
        """
        Return a dictionary of all JETSET definitions in the input parsed file,
        of the form
        "JetSetPar <PARAM>(<PNUMBER>)=<NUMBER>"
        as {'PARAM1': {PNUMBER1: VALUE1, PNUMBER2: VALUE2, ...},
            'PARAM2': {...},
            ...}.
        """
        return get_jetset_definitions(self._parsed_dec_file)

    def list_lineshape_definitions(self):
        """
        Return a dictionary of all SetLineshapePW definitions in the input parsed file,
        of the form
        "SetLineshapePW <MOTHER> <DAUGHTER1> <DAUGHTER2> <VALUE>",
        as
        [(['MOTHER1', 'DAUGHTER1-1', 'DAUGHTER1-2'], VALUE1),
        (['MOTHER2', 'DAUGHTER2-1', 'DAUGHTER2-2'], VALUE2),
        ...]
        """
        return get_lineshape_definitions(self._parsed_dec_file)

    def global_photos_flag(self):
        """
        Return a boolean-like PhotosEnum enum specifying whether or not PHOTOS
        has been enabled.

        Note: PHOTOS is turned on(off) for all decays with the global flag
        yesPhotos(noPhotos).

        Returns
        -------
        out: PhotosEnum, default=PhotosEnum.no
            PhotosEnum.yes / PhotosEnum.no if PHOTOS enabled / disabled
        """
        return get_global_photos_flag(self._parsed_dec_file)

    def list_charge_conjugate_decays(self):
        """
        Return a (sorted) list of all charge conjugate decay definitions
        in the input parsed file, of the form "CDecay <MOTHER>", as
        ['MOTHER1', 'MOTHER2', ...].
        """
        return get_charge_conjugate_decays(self._parsed_dec_file)

    def _find_parsed_decays(self):
        """
        Return a tuple of all decay definitions in the input parsed file,
        of the form
        "Decay <MOTHER>",
        as a tuple of Lark Tree instances with Tree.data=='decay', i.e.,
        (Tree(decay, [Tree(particle, [Token(LABEL, <MOTHER1>]), ...),
        Tree(decay, [Tree(particle, [Token(LABEL, <MOTHER2>]), ...)).

        Duplicate definitions (a bug, of course) are removed, issuing a warning.

        Note
        ----
        1) Method not meant to be used directly!
        2) charge conjugates need to be dealt with differently,
        see 'self._add_charge_conjugate_decays()'.
        """
        if self._parsed_dec_file is not None:
            self._parsed_decays = get_decays(self._parsed_dec_file)

        # Check for duplicates - should be considered a bug in the .dec file!
        self._check_parsed_decays()

    def _add_charge_conjugate_decays(self):
        """
        If requested (see the 'self._include_ccdecays' class attribute),
        create the Lark Tree instances describing the charge conjugate decays
        specified in the input parsed file via the statements of the form
        "CDecay <MOTHER>".
        These are added to the internal list of decays stored in the class
        in variable 'self._parsed_decays', performing a charge conjugate (CC)
        transformation on each CC-related decay, which is cloned.

        Note
        ----
        1) If a decay file only defines 'Decay' decays and no 'CDecay',
        then no charge conjugate decays will be created!
        This seems correct given the "instructions" in the decay file:
        - There is no 'CDecay' statement related to a 'Decay' statement
          for a self-conjugate decaying particle such as the pi0.
        - Else the decay file should be considered incomplete, hence buggy.
        2) Method not meant to be used directly!
        """

        # Do not add any charge conjugate decays if the input parsed file
        # does not define any!
        mother_names_ccdecays = self.list_charge_conjugate_decays()
        if len(mother_names_ccdecays) == 0:
            return

        # Cross-check - make sure charge conjugate decays are not defined
        # with both 'Decay' and 'CDecay' statements!
        mother_names_decays = [get_decay_mother_name(tree)
                               for tree in self._parsed_decays]

        duplicates = [n for n in mother_names_ccdecays if n in mother_names_decays]
        if len(duplicates) > 0:
            msg = """The following particles are defined in the input .dec file with both 'Decay' and 'CDecay': {0}!
The 'CDecay' definition(s) will be ignored ...""".format(', '.join(d for d in duplicates))
            warnings.warn(msg)

        # If that's the case, proceed using the decay definitions specified
        # via the 'Decay' statement, hence discard/remove the definition
        # via the 'CDecay' statement.
        for d in duplicates:
            mother_names_ccdecays.remove(d)

        # We're done if there are no more 'CDecay' decays to treat!
        if len(mother_names_ccdecays) == 0:
            return

        # At last, create the charge conjugate decays:
        # First, make a (deep) copy of the list of relevant Tree instances.
        # Example:
        # if mother_names_ccdecays = ['anti-M10', 'anti-M2+'],
        # the relevant Trees are the ones describing the decays of ['M10', 'M2-'].
        dict_cc_names = self.dict_charge_conjugates()

        # match name -> position in list self._parsed_decays
        name2treepos = {t.children[0].children[0].value:i for i, t in enumerate(self._parsed_decays)}

        trees_to_conjugate = []
        misses = []
        for ccname in mother_names_ccdecays:
            name = find_charge_conjugate_match(ccname, dict_cc_names)
            try:
                match = self._parsed_decays[name2treepos[name]]
                trees_to_conjugate.append(match)
            except:
                misses.append(ccname)
        if len(misses) > 0:
            msg = """\nCorresponding 'Decay' statement for 'CDecay' statement(s) of following particle(s) not found:\n{0}.
Skipping creation of these charge-conjugate decay trees.""".format('\n'.join([m for m in misses]))
            warnings.warn(msg)

        cdecays = [ tree.__deepcopy__(None) for tree in trees_to_conjugate]

        # Finally, perform all particle -> anti(particle) replacements,
        # taking care of charge conjugate decays defined via aliases,
        # passing them as charge conjugates to be processed manually.
        def _is_not_self_conj(t):
            try:
                mname = t.children[0].children[0].value
                if Particle.from_evtgen_name(mname).is_self_conjugate:
                    msg = """Found 'CDecay' statement for self-conjugate particle {0}. This is a bug!
Skipping creation of charge-conjugate decay Tree.""".format(mname)
                    warnings.warn(msg)
                    return False
                else:
                    return True
            except:
                return True
        [ChargeConjugateReplacement(charge_conj_defs=dict_cc_names).visit(t)
        for t in cdecays if _is_not_self_conj(t)]

        # ... and add all these charge-conjugate decays to the list of decays!
        self._parsed_decays.extend(cdecays)

    def _check_parsing(self):
        """Has the .parse() method been called already?"""
        if self._parsed_dec_file is None:
            raise DecFileNotParsed("Hint: call 'parse()'!")

    def _check_parsed_decays(self):
        """
        Is the number of decays parsed consistent with the number of
        decay mother names? An inconsistency can arise if decays are redefined.

        Duplicates are removed, starting from the second occurrence.
        """
        # Issue a helpful warning if duplicates are found
        lmn = self.list_decay_mother_names()
        duplicates = []
        if self.number_of_decays != len(set(lmn)):
            duplicates = set([n for n in lmn if lmn.count(n)>1])
            msg = """The following particle(s) is(are) redefined in the input .dec file with 'Decay': {0}!
All but the first occurrence will be discarded/removed ...""".format(', '.join(d for d in duplicates))
            warnings.warn(msg)

        # Create a list with all occurrences to remove
        # (duplications means multiple instances to remove)
        duplicates_to_remove = []
        for item in duplicates:
            c = lmn.count(item)
            if c>1:
                duplicates_to_remove.extend([item]*(c-1))

        # Actually remove all but the first occurrence of duplicate decays
        for tree in reversed(self._parsed_decays):
            val = tree.children[0].children[0].value
            if val in duplicates_to_remove:
                duplicates_to_remove.remove(val)
                self._parsed_decays.remove(tree)

    @property
    def number_of_decays(self):
        """Return the number of particle decays defined in the parsed .dec file."""
        self._check_parsing()

        return len(self._parsed_decays)

    def list_decay_mother_names(self):
        """
        Return a list of all decay mother names found in the parsed decay file.
        """
        self._check_parsing()

        return [get_decay_mother_name(d) for d in self._parsed_decays]

    def _find_decay_modes(self, mother):
        """
        Return a tuple of Lark Tree instances describing all the decay modes
        of the input mother particle as defined in the parsed .dec file.

        Parameters
        ----------
        mother: str
            Input mother particle name.
        """
        self._check_parsing()

        for decay_Tree in self._parsed_decays:
            if get_decay_mother_name(decay_Tree) == mother:
                return tuple(decay_Tree.find_data('decayline'))

        raise DecayNotFound("Decays of particle '%s' not found in .dec file!" % mother)

    def list_decay_modes(self, mother, pdg_name=False):
        """
        Return a list of decay modes for the given mother particle.

        Parameters
        ----------
        mother: str
            Input mother particle name.
        pdg_name: str, optional, default=False
            Input mother particle name is the PDG name,
            not the (default) EvtGen name.

        Example
        -------
        >>> parser = DecFileParser('my-decay-file.dec')
        >>> parser.parse()
        >>> parser.list_decay_mother_names()  # Inspect what decays are defined
        >>> parser.list_decay_modes('pi0')
        """
        if pdg_name:
            mother = PDG2EvtGenNameMap[mother]

        return [get_final_state_particle_names(mode)
                for mode in self._find_decay_modes(mother)]

    def _decay_mode_details(self, decay_mode):
        """
        Parse a decay mode (Tree instance)
        and return the relevant bits of information in it.
        """

        bf = get_branching_fraction(decay_mode)
        fsp_names = get_final_state_particle_names(decay_mode)
        model = get_model_name(decay_mode)
        model_params = get_model_parameters(decay_mode)

        return (bf, fsp_names, model, model_params)

    def print_decay_modes(self, mother,
                          pdg_name=False,
                          print_model=True,
                          ascending=False):
        """
        Pretty print of the decay modes of a given particle.

        Parameters
        ----------
        mother: str
            Input mother particle name.
        pdg_name: str, optional, default=False
            Input mother particle name is the PDG name,
            not the (default) EvtGen name.
        print_model: bool, optional, default=True
            Specify whether to print the decay model and model parameters,
            if available.
        ascending: bool, optional, default=False
            Print the list of decay modes ordered in ascending/descending order
            of branching fraction.
        """
        if pdg_name:
            mother = PDG2EvtGenNameMap[mother]

        dms = self._find_decay_modes(mother)

        l = list()
        if print_model:
            for dm in dms:
                dm_details = self._decay_mode_details(dm)
                l.append((dm_details[0],'%-50s %15s %s' % ('  '.\
                    join(p for p in dm_details[1]), dm_details[2], dm_details[3])))
        else:
            for dm in dms:
                fsp_names = get_final_state_particle_names(dm)
                l.append((get_branching_fraction(dm),
                         '%-50s' % ('  '.join(p for p in fsp_names))))

        l.sort(key=operator.itemgetter(0), reverse=(not ascending))

        for bf, info in l:
            print('%12g : %s' % (bf, info))

    def build_decay_chains(self, mother, stable_particles=[]):
        """
        Iteratively build the entire decay chains of a given mother particle,
        optionally considering, on the fly, certain particles as stable.
        This way, for example, only the B -> D E F part in a decay chain
        A -> B (-> D E F (-> G H)) C
        can be trivially selected for inspection.

        Parameters
        ----------
        mother: str
            Input mother particle name.
        stable_particles: iterable, optional, default=[]
            If provided, stops the decay-chain parsing,
            taking the "list" as particles to be considered stable.

        Returns
        -------
        out: dict
            Decay chain as a dictionary of the form
            {mother: [{'bf': float, 'fs': list, 'model': str, 'model_params': str}]}
            where
            'bf' stands for the deca mode branching fraction,
            'fs' is a list of final-state particle names (strings)
            and/or dictionaries of the same form as the decay chain above,
            'model' is the model name, if found, else '',
            'model_params' are the model parameters, if specified, else ''

        Examples
        --------
        >>> parser = DecFileParser('a-Dplus-decay-file.dec')
        >>> parser.parse()
        >>> parser.build_decay_chains('D+')
        {'D+': [{'bf': 1.0,
           'fs': ['K-',
            'pi+',
            'pi+',
            {'pi0': [{'bf': 0.988228297,
               'fs': ['gamma', 'gamma'],
               'model': 'PHSP',
               'model_params': ''},
              {'bf': 0.011738247,
               'fs': ['e+', 'e-', 'gamma'],
               'model': 'PI0_DALITZ',
               'model_params': ''},
              {'bf': 3.3392e-05,
              'fs': ['e+', 'e+', 'e-', 'e-'],
              'model': 'PHSP',
              'model_params': ''},
              {'bf': 6.5e-08, 'fs': ['e+', 'e-'], 'model': 'PHSP', 'model_params': ''}]}],
           'model': 'PHSP',
           'model_params': ''}]}
        >>> p.build_decay_chains('D+', stable_particles=['pi0'])
        {'D+': [{'bf': 1.0, 'fs': ['K-', 'pi+', 'pi+', 'pi0'], 'model': 'PHSP', 'model_params': ''}]}
        """
        keys = ('bf', 'fs', 'model', 'model_params')

        info = list()
        for dm in (self._find_decay_modes(mother)):
            list_dm_details = self._decay_mode_details(dm)
            d = dict(zip(keys,list_dm_details))

            for i, fs in enumerate(d['fs']):
                if fs in stable_particles:
                    continue

                try:
                    # This throws a DecayNotFound exception
                    # if fs does not have decays defined in the parsed file
                    _n_dms = len(self._find_decay_modes(fs))

                    _info = self.build_decay_chains(fs, stable_particles)
                    d['fs'][i] = _info
                except DecayNotFound:
                    pass

            info.append(d)

        info = {mother:info}
        return info

    def __repr__(self):
        if self._parsed_dec_file is not None:
            return "<{self.__class__.__name__}: decfile(s)={decfile}, n_decays={n_decays}>".format(
                self=self, decfile=self._dec_file_names, n_decays=self.number_of_decays)
        else:
            return "<{self.__class__.__name__}: decfile(s)={decfile}>"\
                    .format(self=self, decfile=self._dec_file_names)

    def __str__(self):
        return repr(self)


class DecayModelParamValueReplacement(Visitor):
    """
    Lark Visitor implementing the replacement of decay model parameter names
    with the actual parameter values provided in 'Define' statements,
    and replacement of floats stored as strings to the actual floating values.
    The replacement is only relevant for Lark Tree instances of name
    'model_options' (Tree.data == 'model_options').

    Parameters
    ----------
    define_defs: dict, optional, default={}
        Dictionary with the 'Define' definitions in the parsed file.
        Argument to be passed to the class constructor.

    Examples
    --------
    >>> from lark import Tree, Token
    >>> ...
    >>> t = Tree('decay', [Tree('particle', [Token('LABEL', 'Upsilon(4S)')]),
    ...         Tree('decayline', [Tree('value', [Token('SIGNED_NUMBER', '1.0')]),
    ...         Tree('particle', [Token('LABEL', 'B0')]),
    ...         Tree('particle', [Token('LABEL', 'anti-B0')]),
    ...         Tree('model', [Token('MODEL_NAME', 'VSS_BMIX'),
    ...         Tree('model_options', [Token('LABEL', 'dm')])])])])
    >>> dict_define_defs = {'dm': 0.507e12}
    >>> DecayModelParamValueReplacement(define_defs=dict_define_defs).visit(t)
    Tree(decay, [Tree(particle, [Token(LABEL, 'Upsilon(4S)')]), Tree(decayline,
    [Tree(value, [Token(SIGNED_NUMBER, '1.0')]), Tree(particle, [Token(LABEL, 'B0')]),
    Tree(particle, [Token(LABEL, 'anti-B0')]), Tree(model, [Token(MODEL_NAME, 'VSS_BMIX'),
    Tree(model_options, [Token(LABEL, 507000000000.0)])])])])
    """
    def __init__(self, define_defs=dict()):
        self.define_defs = define_defs

    def _replacement(self, t):
        try:
            t.children[0].value = float(t.children[0].value)
        except AttributeError:
            if t.value in self.define_defs:
                t.value = self.define_defs[t.value]

    def model_options(self, tree):
        """
        Method for the rule (here, a replacement) we wish to implement.
        """
        assert tree.data == 'model_options'

        for child in tree.children:
            self._replacement(child)


class ChargeConjugateReplacement(Visitor):
    """
    Lark Visitor implementing the replacement of all particle names
    with their charge conjugate particle names
    in a Lark Tree of name 'particle' (Tree.data == 'particle').

    Note
    ----
    1) There is no check of whether the mother particle is self-conjugate or not.
    It is the responsibility of the caller to make sure the operation
    is not trivial, meaning returning the same (self-conjugate) decay!
    2) If a particle name (say, 'UNKOWN') is not found or known,
    (search done via the Particle class in the particle package),
    its charge conjugate name is denoted as 'ChargeConj(UNKOWN)'.

    Parameters
    ----------
    charge_conj_defs: dict, optional, default={}
        Dictionary with the charge conjugate particle definitions
        in the parsed file. Argument to be passed to the class constructor.

    Examples
    --------
    >>> from lark import Tree, Token
    >>> ...
    >>> t = Tree('decay', [Tree('particle', [Token('LABEL', 'D0')]), Tree('decayline', [Tree
    ... ('value', [Token('SIGNED_NUMBER', '1.0')]), Tree('particle', [Token('LABEL', 'K-')])
    ... , Tree('particle', [Token('LABEL', 'pi+')]), Tree('model', [Token('MODEL_NAME', 'PHS
    ... P')])])])
    >>> ChargeConjugateReplacement().visit(t)
    Tree(decay, [Tree(particle, [Token(LABEL, 'D~0')]), Tree(decayline,
    [Tree(value, [Token(SIGNED_NUMBER, '1.0')]), Tree(particle, [Token(LABEL, 'K+')]),
    Tree(particle, [Token(LABEL, 'pi-')]), Tree(model, [Token(MODEL_NAME, 'PHSP')])])])
    """
    def __init__(self, charge_conj_defs=dict()):
        self.charge_conj_defs = charge_conj_defs

    def particle(self, tree):
        """
        Method for the rule (here, a replacement) we wish to implement.
        """
        assert tree.data == 'particle'
        pname = tree.children[0].value
        ccpname = find_charge_conjugate_match(pname, self.charge_conj_defs)
        self.charge_conj_defs[pname] = ccpname
        tree.children[0].value = ccpname

def find_charge_conjugate_match(pname, dict_cc_names=dict()):
    """
    Find the charge-conjugate particle name making use of user information
    from "ChargeConj" statements in a decay file.

    The name `ChargeConj(pname)` is returned if all matching "routes" fail,
    see `charge_conjugate_name(...) function.`
    """
    # Check the list of particle-antiparticle matches provided ;-)
    if len(dict_cc_names) > 0:
        match = dict_cc_names.get(pname)
        if match is not None:
            return match
        # Yes, both 'ChargeConj P CCP' and 'ChargeConj CCP P' are relevant
        for p, ccp in dict_cc_names.items():
            if ccp == pname:
                return p

    return charge_conjugate_name(pname)

def get_decay_mother_name(decay_tree):
    """
    Return the mother particle name for the decay mode defined
    in the input Tree of name 'decay'.

    Parameters
    ----------
    decay_tree: Lark Tree instance
        Input Tree satisfying Tree.data=='decay'.
    """
    if not isinstance(decay_tree, Tree) or decay_tree.data != 'decay':
        raise RuntimeError("Input not an instance of a 'decay' Tree!")

    # For a 'decay' Tree, tree.children[0] is the mother particle Tree
    # and tree.children[0].children[0].value is the mother particle name
    return decay_tree.children[0].children[0].value


def get_branching_fraction(decay_mode):
    """
    Return the branching fraction (float) for the decay mode defined
    in the input Tree of name 'decayline'.

    Parameters
    ----------
    decay_mode: Lark Tree instance
        Input Tree satisfying Tree.data=='decayline'.
    """
    if not isinstance(decay_mode, Tree) or decay_mode.data != 'decayline':
        raise RuntimeError("Check your input, not an instance of a 'decayline' Tree!")

    # For a 'decayline' Tree, Tree.children[0] is the branching fraction Tree
    # and tree.children[0].children[0].value is the BF stored as a str
    try:  # the branching fraction value as a float
        return float(decay_mode.children[0].children[0].value)
    except RuntimeError:
        raise RuntimeError("'decayline' Tree does not seem to have the usual structure. Check it.")


def get_final_state_particles(decay_mode):
    """
    Return a list of Lark Tree instances describing the final-state particles
    for the decay mode defined in the input Tree of name 'decayline'.

    Parameters
    ----------
    decay_mode: Lark Tree instance
        Input Tree satisfying Tree.data=='decayline'.

    Examples
    --------
    For a decay
        Decay MyB_s0
            1.000     K+     K-     SSD_CP 20.e12 0.1 1.0 0.04 9.6 -0.8 8.4 -0.6;
        Enddecay
    the list
        [Tree(particle, [Token(LABEL, 'K+')]), Tree(particle, [Token(LABEL, 'K-')])]
    will be returned.
    """
    if not isinstance(decay_mode, Tree) or decay_mode.data != 'decayline':
        raise RuntimeError("Input not an instance of a 'decayline' Tree!")

    # list of Trees of final-state particles
    return list(decay_mode.find_data('particle'))


def get_final_state_particle_names(decay_mode):
    """
    Return a list of final-state particle names for the decay mode defined
    in the input Tree of name 'decayline'.

    Parameters
    ----------
    decay_mode: Lark Tree instance
        Input Tree satisfying Tree.data=='decayline'.

    Examples
    --------
    For a decay
        Decay MyB_s0
            1.000     K+     K-     SSD_CP 20.e12 0.1 1.0 0.04 9.6 -0.8 8.4 -0.6;
        Enddecay
    the list ['K+', 'K-'] will be returned.
    """
    if not isinstance(decay_mode, Tree) or decay_mode.data != 'decayline':
        raise RuntimeError("Input not an instance of a 'decayline' Tree!")

    fsps = get_final_state_particles(decay_mode)
    # list of final-state particle names
    return [str(fsp.children[0].value) for fsp in fsps]


def get_model_name(decay_mode):
    """
    Return the decay model name in a Tree of name 'decayline'.

    Parameters
    ----------
    decay_mode: Lark Tree instance
        Input Tree satisfying Tree.data=='decayline'.

    Examples
    --------
    For a decay
        Decay MyB_s0
            1.000     K+     K-     SSD_CP 20.e12 0.1 1.0 0.04 9.6 -0.8 8.4 -0.6;
        Enddecay
    the string 'SSD_CP' will be returned.
    """
    if not isinstance(decay_mode, Tree) or decay_mode.data != 'decayline':
        raise RuntimeError("Input not an instance of a 'decayline' Tree!")

    lm = list(decay_mode.find_data('model'))
    return str(lm[0].children[0].value)


def get_model_parameters(decay_mode):
    """
    Return a list of decay model parameters in a Tree of name 'decayline',
    if defined, else an empty string.

    Parameters
    ----------
    decay_mode: Lark Tree instance
        Input Tree satisfying Tree.data=='decayline'.

    Examples
    --------
    For a decay
        Decay MyB_s0
            1.000     K+     K-     SSD_CP 20.e12 0.1 1.0 0.04 9.6 -0.8 8.4 -0.6;
        Enddecay
    the list
        [20000000000000.0, 0.1, 1.0, 0.04, 9.6, -0.8, 8.4, -0.6]
    will be returned.

    For a decay
        Decay MyD0
            1.00      K-   pi-   pi+   pi+     LbAmpGen DtoKpipipi_v1 ;
        Enddecay
    the list
        ['DtoKpipipi_v1']
    will be returned.
    """
    if not isinstance(decay_mode, Tree) or decay_mode.data != 'decayline':
        raise RuntimeError("Input not an instance of a 'decayline' Tree!")

    lmo = list(decay_mode.find_data('model_options'))

    def _value(t):
        try:
            return t.children[0].value
        except AttributeError:
            return t.value

    return [_value(tree) for tree in lmo[0].children] if len(lmo) == 1 else ''


def get_decays(parsed_file):
    """
    Return a list of all decay definitions in the input parsed file,
    of the form
    "Decay <MOTHER>",
    as a tuple of Lark Tree instances with Tree.data=='decay', i.e.,
    [Tree(decay, [Tree(particle, [Token(LABEL, <MOTHER1>]), ...),
     Tree(decay, [Tree(particle, [Token(LABEL, <MOTHER2>]), ...)].

    Parameters
    ----------
    parsed_file: Lark Tree instance
        Input parsed file.
    """
    if not isinstance(parsed_file, Tree) :
        raise RuntimeError("Input not an instance of a Tree!")

    try:
        return list(parsed_file.find_data('decay'))
    except:
        RuntimeError("Input parsed file does not seem to have the expected structure.")


def get_charge_conjugate_decays(parsed_file):
    """
    Return a (sorted) list of all charge conjugate decay definitions
    in the input parsed file, of the form "CDecay <MOTHER>", as
    ['MOTHER1', 'MOTHER2', ...].

    Parameters
    ----------
    parsed_file: Lark Tree instance
        Input parsed file.
    """
    if not isinstance(parsed_file, Tree) :
        raise RuntimeError("Input not an instance of a Tree!")

    try:
        return sorted([tree.children[0].children[0].value for tree in parsed_file.find_data('cdecay')])
    except:
        RuntimeError("Input parsed file does not seem to have the expected structure.")


def get_definitions(parsed_file):
    """
    Return a dictionary of all definitions in the input parsed file, of the form
    "Define <NAME> <VALUE>", as {'NAME1': VALUE1, 'NAME2': VALUE2, ...}.

    Parameters
    ----------
    parsed_file: Lark Tree instance
        Input parsed file.
    """
    if not isinstance(parsed_file, Tree) :
        raise RuntimeError("Input not an instance of a Tree!")

    try:
        return {tree.children[0].children[0].value:float(tree.children[1].children[0].value)
            for tree in parsed_file.find_data('define')}
    except:
        RuntimeError("Input parsed file does not seem to have the expected structure.")


def get_aliases(parsed_file):
    """
    Return a dictionary of all aliases in the input parsed file, of the form
    "Alias <NAME> <ALIAS>", as {'NAME1': ALIAS1, 'NAME2': ALIAS2, ...}.

    Parameters
    ----------
    parsed_file: Lark Tree instance
        Input parsed file.
    """
    if not isinstance(parsed_file, Tree) :
        raise RuntimeError("Input not an instance of a Tree!")

    try:
        return {tree.children[0].children[0].value:tree.children[1].children[0].value
            for tree in parsed_file.find_data('alias')}
    except:
        RuntimeError("Input parsed file does not seem to have the expected structure.")


def get_charge_conjugate_defs(parsed_file):
    """
    Return a dictionary of all charge conjugate definitions
    in the input parsed file, of the form "ChargeConj <PARTICLE> <CC_PARTICLE>",
    as {'PARTICLE1': CC_PARTICLE1, 'PARTICLE2': CC_PARTICLE2, ...}.

    Parameters
    ----------
    parsed_file: Lark Tree instance
        Input parsed file.
    """
    if not isinstance(parsed_file, Tree) :
        raise RuntimeError("Input not an instance of a Tree!")

    try:
        return {tree.children[0].children[0].value:tree.children[1].children[0].value
            for tree in parsed_file.find_data('chargeconj')}
    except:
        RuntimeError("Input parsed file does not seem to have the expected structure.")


def get_pythia_definitions(parsed_file):
    """
    Return a dictionary of all Pythia definitions in the input parsed file,
    of the form
    "PythiaBothParam <NAME>=<LABEL>"
    or
    "PythiaBothParam <NAME>=<NUMBER>",
    as {'NAME1': 'LABEL1', 'NAME2': VALUE2, ...}.

    Parameters
    ----------
    parsed_file: Lark Tree instance
        Input parsed file.
    """
    if not isinstance(parsed_file, Tree) :
        raise RuntimeError("Input not an instance of a Tree!")

    def str_or_float(arg):
        try:
            return float(arg)
        except:
            return arg

    try:
        return {'{0}:{1}'.format(tree.children[0].value, tree.children[1].value):str_or_float(tree.children[2].value)
            for tree in parsed_file.find_data('pythia_def')}
    except:
        RuntimeError("Input parsed file does not seem to have the expected structure.")


def get_jetset_definitions(parsed_file):
    """
    Return a dictionary of all JETSET definitions in the input parsed file,
    of the form
    "JetSetPar <PARAM>(<PNUMBER>)=<NUMBER>"
    as {'PARAM1': {PNUMBER1: VALUE1, PNUMBER2: VALUE2, ...},
        'PARAM2': {...},
        ...}.

    Parameters
    ----------
    parsed_file: Lark Tree instance
        Input parsed file.
    """
    if not isinstance(parsed_file, Tree) :
        raise RuntimeError("Input not an instance of a Tree!")

    get_jetsetpar = re.compile(r"""
    ^                                     # Beginning of string
        (?P<pname>      [a-zA-Z]+?  )     # One or more characters, non-greedy
     \( (?P<pnumber>    \d+         ) \)  # parameter number in ()
    """, re.VERBOSE)

    def to_int_or_float(n):
        """
        Trivial helper function to convert the parsed (as strings)
        JETSET parameters into what they are, namely integers or floats.
        """
        try:
            return int(n)
        except ValueError:
            try:
                return float(n)
            except:
                # pass though non-numbers unchanged
                return n

    try:
        dict_params = {}
        for tree in parsed_file.find_data('jetset_def'):
            param = get_jetsetpar.match(tree.children[0].value).groupdict()
            try:
                dict_params[param['pname']].update(
                    {int(param['pnumber']):to_int_or_float(tree.children[1].value)})
            except KeyError:
                dict_params[param['pname']] =\
                    {int(param['pnumber']):to_int_or_float(tree.children[1].value)}
        return dict_params
    except:
        RuntimeError("Input parsed file does not seem to have the expected structure.")


def get_lineshape_definitions(parsed_file):
    """
    Return a dictionary of all SetLineshapePW definitions in the input parsed file,
    of the form
    "SetLineshapePW <MOTHER> <DAUGHTER1> <DAUGHTER2> <VALUE>",
    as
    [(['MOTHER1', 'DAUGHTER1-1', 'DAUGHTER1-2'], VALUE1),
     (['MOTHER2', 'DAUGHTER2-1', 'DAUGHTER2-2'], VALUE2),
     ...]

    Parameters
    ----------
    parsed_file: Lark Tree instance
        Input parsed file.
    """
    if not isinstance(parsed_file, Tree) :
        raise RuntimeError("Input not an instance of a Tree!")

    try:
        d = list()
        for tree in parsed_file.find_data('setlspw'):
            particles = [p.children[0].value for p in tree.children[:-1]]
            val = int(tree.children[3].children[0].value)
            d.append((particles, val))
        return d
    except:
        RuntimeError("Input parsed file does not seem to have the expected structure.")


def get_global_photos_flag(parsed_file):
    """
    Return a boolean-like PhotosEnum enum specifying whether or not PHOTOS
    has been enabled.

    Note: PHOTOS is turned on(off) for all decays with the global flag
    yesPhotos(noPhotos).

    Parameters
    ----------
    parsed_file: Lark Tree instance
        Input parsed file.

    Returns
    -------
    out: PhotosEnum, default=PhotosEnum.no
        PhotosEnum.yes / PhotosEnum.no if PHOTOS enabled / disabled
    """
    if not isinstance(parsed_file, Tree) :
        raise RuntimeError("Input not an instance of a Tree!")

    # Check if the flag is not set more than once, just in case ...
    tree = tuple(parsed_file.find_data('global_photos'))
    if len(tree) == 0:
        return PhotosEnum.no
    elif len(tree) > 1:
            warnings.warn("PHOTOS flag re-set! Using flag set in last ...")

    tree = tree[-1]

    try:
        val = tree.children[0].data
        return PhotosEnum.yes if val=='yes' else PhotosEnum.no
    except:
        RuntimeError("Input parsed file does not seem to have the expected structure.")


class TreeToDec(Transformer):
    def yes(self, items):
        return True
    def no(self, items):
        return False
    def global_photos(self, items):
        item, = items
        return PhotosEnum.yes if item else PhotosEnum.no
    def value(self, items):
        item, = items
        return float(item)
    def label(self, items):
        item, = items
        return str(item)
    def photos(self, items):
        return PhotosEnum.yes


def define(transformed):
    return {x.children[0]:x.children[1] for x in transformed.find_data('define')}
def pythia_def(transformed):
    return [x.children for x in transformed.find_data('pythia_def')]
def alias(transformed):
    return {x.children[0]:x.children[1] for x in transformed.find_data('alias')}

def chargeconj(transformed):
    return {x.children[0]:x.children[1] for x in transformed.find_data('chargeconj')}

# Commands
def global_photos(transformed):
    return {x.children[0]:x.children[1] for x in transformed.find_data('global_photos')}

def decay(transformed):
    return Tree('decay', list(transformed.find_data('decay')))
def cdecay(transformed):
    return [x.children[0] for x in transformed.find_data('cdecay')]
def setlspw(transformed):
    return list(transformed.find_data('setlspw'))



# Copyright (c) 2018-2020, Eduardo Rodrigues and Henry Schreiner.
#
# Distributed under the 3-clause BSD license, see accompanying file LICENSE
# or https://github.com/scikit-hep/decaylanguage for details.

'''
A general base class representing decays.
'''


from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import attr
import warnings
from itertools import product

from particle import Particle

from ..utils import iter_flatten


try:
    import graphviz
except ImportError:
    graphviz = None
    warnings.warn("Graphviz is not installed. Line display not available.")

@attr.s(slots=True)
class ModelDecay(object):
    '''
    This describes a decay very generally, with search and print features.
    Subclassed for futher usage.
    '''

    particle = attr.ib()
    daughters = attr.ib([], converter=lambda x: x if x else [])
    name = attr.ib(None)

    def __attrs_post_init__(self):
        if self.name is None:
            self.name = self.particle.name

    def is_vertex(self):
        return len(self) == 2

    def is_strong(self):
        if not self.is_vertex():
            return None
        return set(self.particle.quarks) == set(self[0].particle.quarks).union(set(self[1].particle.quarks))

    def __len__(self):
        return len(self.daughters)

    def __getitem__(self, item):
        return self.daughters[item]

    def _get_html(self):
        '''
        Get the html dot representation of this node only. Override in subclasses.
        '''
        return self.particle.html_name

    def _add_nodes(self, drawing):
        name = self._get_html()
        drawing.node(str(id(self)), "<" + name + ">")
        for p in self.daughters:
            drawing.edge(str(id(self)), str(id(p)))
            p._add_nodes(drawing)


    @property
    def vertexes(self):
        verts = []
        for d in self.daughters:
            if d.is_vertex():
                verts.append(d)
                verts += d.vertexes
        return verts

    @property
    def structure(self):
        '''
        The structure of the decay chain, simplified to only final state particles
        '''
        if self.daughters:
            return [d.structure for d in self.daughters]
        else:
            return self.particle

    def list_structure(self, final_states):
        '''
        The structure in the form [(0,1,2,3)], where the dual-list is used
        for permutations for bose symmatrization.
        So for final_states=[a,b,c,c], [a,c,[c,b]] would be:
        [(0,2,3,1),(0,3,2,1)]
        '''

        structure = list(iter_flatten(self.structure))

        if set(structure) - set(final_states):
            raise RuntimeError("The final states must encompass all particles in final states!")

        possibilities = [[i for i, v in enumerate(final_states) if v == name] for name in structure]
        return [a for a in product(*possibilities) if len(set(a)) == len(a)]

    def __str__(self):
        name = str(self.particle)

        if self.daughters:
            name += '{'+','.join(map(str, self.daughters))+'}'
        return name

    if graphviz:
        def _make_graphviz(self):
            d = graphviz.Digraph()
            d.attr(labelloc='t', label=str(self))
            self._add_nodes(d)
            return d

        def _repr_svg_(self):
            return self._make_graphviz()._repr_svg_()



# Copyright (c) 2018-2020, Eduardo Rodrigues and Henry Schreiner.
#
# Distributed under the 3-clause BSD license, see accompanying file LICENSE
# or https://github.com/scikit-hep/decaylanguage for details.

"""
Collection of enums and info to help characterising .dec decay files.
"""

# Backport needed if Python 2 is used
from enum import IntEnum


class PhotosEnum(IntEnum):
    no = 0
    yes = 1


# This list should match the list specified in the decay file parser file
# 'decaylanguage/data/decfile.lark'!
known_decay_models = (
    "BaryonPCR",
    "BCL",
    "BGL",
    "BTO3PI_CP",
    "BTOSLLALI",
    "BTOSLLBALL",
    "BTOXSGAMMA",
    "BTOXSLL",
    "CB3PI-MPP",
    "CB3PI-P00",
    "D_DALITZ",
    "ETAPRIME_DALITZ",
    "ETA_DALITZ",
    "ETA_FULLDALITZ",
    "ETA_PI0DALITZ",
    "GOITY_ROBERTS",
    "HELAMP",
    "HQET",
    "ISGW2",
    "LbAmpGen",
    "LLSW",
    "OMEGA_DALITZ",
    "PARTWAVE",
    "PHSP",
    "PI0_DALITZ",
    "PYCONT",
    "PYTHIA",
    "SLN",
    "SSD_CP",
    "SSS_CP",
    "STS",
    "SVP_HELAMP",
    "SVS",
    "SVV_CP",
    "SVV_HELAMP",
    "TAUHADNU",
    "TAULNUNU",
    "TAUSCALARNU",
    "TAUVECTORNU",
    "TSS",
    "TVS_PWAVE",
    "VLL",
    "VSP_PWAVE",
    "VSS_BMIX",
    "VSS",
    "VUB",
    "VVPIPI",
    "VVP",
    "VVS_PWAVE",
    "YMSTOYNSPIPICLEO"
    )



# Copyright (c) 2018-2020, Eduardo Rodrigues and Henry Schreiner.
#
# Distributed under the 3-clause BSD license, see accompanying file LICENSE
# or https://github.com/scikit-hep/decaylanguage for details.


__version__ = '0.7.2'

version = __version__
version_info = __version__.split('.')



#!/usr/bin/env python

from __future__ import print_function, division, absolute_import

#try:
#    import cmd2 as cmd
#except:
import cmd
import os

from particle import Particle


DIR = os.path.dirname(__file__)

known_models = ['VSS', 'VSS_BMIX', 'PHSP', 'HELAMP', 'JETSET', 'PHOTOS',
                'ISGW2', 'HQET', 'GOITY_ROBERTS', 'VUB', 'SVS',
                'SVV_HELAMP', 'BTOXSGAMMA', 'BTOSLLBALL', 'BTOXSLL',
                'SSD_CP', 'SVV_CP', 'BTO3PI_CP', 'STS', 'JSCONT', 'SLN',
                'CB3PI-P00', 'CB3PI-MPP', 'VSP_PWAVE', 'TAUSCALARNU',
                'TAUHADNU', 'D_DALITZ', 'VLL', 'TSS', 'VVP', 'VVS_PWAVE',
                'PARTWAVE', 'TVS_PWAVE', 'OMEGA_DALITZ', 'ETA_DALITZ',
                'PI0_DALITZ', 'TAUVECTORNU', 'SVP_HELAMP', 'VVPIPI',
                'VPHOTOV', 'PYTHIA', 'BTOSLLALI'
                ]

import os
if 'C3_DATA' in os.environ:
    defdecfile = '%s/DECAY.DEC' % os.environ['C3_DATA']
else:
    defdecfile = os.path.join(DIR, '../data/DECAY_LHCB.DEC')

class DaughterList(dict):
    def __init__(self):
        dict.__init__(self)

    def add(self, daughter):
        self[daughter] = self.get(daughter, 0)

class AllowedDecays(object):
    def __init__(self, particle):
        self.decay_of = particle
        self.decays = []

class Decay(object):
    def __init__(self, bf = 0, daughters = None):
        if daughters == None:
            daughters = DaughterList()
        self.bf = bf
        self.daughters = daughters

    def daughters_to_string(self):
        strs = []
        for dau in self.daughters:
            strs += [dau]*self.daughters[dau]
        def sortkey(x):
            p = Particle.from_search_list(name=x)
            if len(p) == 1:
                return p[0]
            else:
                return 100000
        strs.sort(key=sortkey, reverse=True)
        return ' '.join(strs)


class decparser(cmd.Cmd):
    def __init__(self, stdin=None, stdout=None):
        self.prompt = ''
        self.file_parse_status = ''
        self.current_decay_top = None
        self.ignoreUntilSemicolon = False
        self.decaylist = {}
        self.cdecay_delayed = []
        cmd.Cmd.__init__(self, 'tab', stdin, stdout)

    def default(self, line):
        if line[:1] == '#' or 'Photos' in line:
            pass
        elif self.ignoreUntilSemicolon:
            if ';' in line:
                self.ignoreUntilSemicolon = False
        elif self.file_parse_status == 'Decay':
            self.addline(line)
        else:
            cmd.Cmd.default(self, line)

    def emptyline(self):
        return ''

##    def precmd(self, line):
##        print line
##        return line

    def do_EOF(self, line):
        return True

    def do_Define(self, line):
        pass

    def do_Alias(self, line):
        pass

    def do_ChargeConj(self, line):
        pass

    def do_JetSetPar(self, line):
        pass

    def do_SetLineshapePW(self, line):
        pass

    def do_ModelAlias(self, line):
        if not ';' in line:
            self.ignoreUntilSemicolon = True

    def do_Decay(self, line):
        if self.file_parse_status == 'Decay':
            raise Exception('Repeated Decay statement: %s' % line)
        else:
            self.file_parse_status = 'Decay'
            particle = Particle.from_string(line.split()[0])
            self.current_decay_top = AllowedDecays(particle)

    def do_CDecay(self, line):
##        print 'CDecay for', line
        if self.file_parse_status == 'Decay':
            raise Exception('Cannot do CDecay in Decay block: %s' % line)
        conj = Particle.from_string(line.split()[0]).invert()
        if conj not in self.decaylist:
            print(line)
            self.cdecay_delayed.append(line)
            return
            raise Exception('CDecay without conjugate mode: %s' % line)
        self.current_decay_top = AllowedDecays(line.split()[0])
##        print conj
        for cdecay in self.decaylist[conj].decays:
##            print cdecay.daughters
            ndecay = Decay()
            ndecay.bf = cdecay.bf
            for dau in cdecay.daughters:
##                print cdecay.daughters
                ndecay.daughters[Particle.from_string(dau).invert()] = cdecay.daughters[dau]
##            print                 ndecay.daughters

##            print cdecay.daughters, ndecay.daughters
##        print line
            self.current_decay_top.decays.append(ndecay)
        self.decaylist[self.current_decay_top.decay_of] = self.current_decay_top
        self.current_decay_top = None

    def do_End(self, line):
##        return self.do_EOF(line)
        if self.cdecay_delayed != []:
            print('Ought to be [] 0!:', self.cdecay_delayed, len(self.cdecay_delayed))
        while self.cdecay_delayed:
            self.do_CDecay(self.cdecay_delayed.pop())
        pass

    def addline(self, line):
        spl = line.split()
        try:
            bf = float(spl[0])
        except:
            raise Exception('Cannot parse decay line: %s' % line)
##        mod_found = False
##         for x in known_models:
##             if x in line:
##                 mod_found = True
##         if not mod_found:
##             print '\n', line
        while not spl[-1][-1:] == ';':
##            print 'fixin', line
            if self.use_rawinput:
                line = ' '.join((line, input()))
            else:
                line = ' '.join((line, self.stdin.readline()))
            spl = line.split()
##        print spl
        killindex = None; i = 0
##        print spl
        while killindex == None and i < len(spl):
            if spl[i][-1] == ';': spl[i] = spl[i][:-1]
            if spl[i] in known_models:
                killindex = i
            i += 1
        if killindex == None:
            print(spl)
            raise Exception('No decay model specified: %s' % line)
##        print spl[killindex:]
        decay = Decay()
        decay.bf = bf
##        print self.current_decay_top.decay_of, 'to',
        for part in spl[1:killindex]:
##            print part,
            if part in decay.daughters:
                decay.daughters[part] += 1
            else:
                decay.daughters[part] = 1
##        print decay.daughters
        self.current_decay_top.decays.append(decay)
##        print self.current_decay_top.decays

    def do_Enddecay(self, line):
        if self.file_parse_status != 'Decay':
            raise Exception('Enddecay with no decay: %s' % line)
        else:
            self.file_parse_status = ''
            self.decaylist[self.current_decay_top.decay_of] = self.current_decay_top
            self.current_decay_top = None



    def do_hi(self, aft):
        print('hi', aft)


class interactive(cmd.Cmd):
    def __init__(self):
        cmd.Cmd.__init__(self)
        self.decaylist = {}
        # These are particles we don't decay
        self.termpart = ['pi0', 'K_S0']

    def do_readfile(self, line):
        "Read a file in (defaults to the one built into this package)"
        if line == '':
            fname = defdecfile
        else:
            fname = line
        with open(fname, 'r') as infile:
            q = decparser(stdin=infile)
            q.use_rawinput = False
            q.cmdloop()
            self.decaylist = q.decaylist

    def do_dump(self, line):
##        print line
        if line == '':
            lpart = self.decaylist
        else:
            lpart = line.split()
        for part in lpart:
            if part not in self.decaylist:
                print('Unknown particle %s' % part)
            else:
                print('-------------', part, '-------------')
                for decay in self.decaylist[part].decays:
                    print(decay.bf, end=' ')
                    print(decay.daughters_to_string(), end=' ')
                    print()


    def do_termpart(self, line):
        if line == '':
            print(self.termpart)
        elif len(line.split()) == 2:
            cmd = line.split()[0].lower()
            if cmd not in ('add', 'del'):
                print('Unknown command %s' % cmd)
            else:
                part = line.split()[1]
                if cmd == 'add' and part not in self.termpart:
                    self.termpart.append(part)
                elif cmd == 'del':
                    if part not in self.termpart:
                        print('%s not in list of terminating particles' % part)
                    else:
                        del self.termpart[self.termpart.index(part)]
        else:
            print('Syntax: termpart [(add|del) particle]')

    def do_exit(self, line):
        return self.do_EOF(line)

    def do_quit(self, line):
        return self.do_EOF(line)

    def do_final(self, line):
##         predeclist = self.decaylist[line].decays[:]
##         declist = []
##         for dec in predeclist:
##             declist.append([[],dec])
##         decaytable = self.decaylist.copy()
##         for j in self.termpart:
##             del decaytable[j]
##         last = []
##         while last != declist:
##             last = declist[:]
##             recurseOneLevel(declist, decaytable)
        declist = self.getDecList(line)
        declist = compactDecayList(declist)
        declist.sort(key=lambda x: x.bf, reverse=True)
        for dec in declist:
##            print dec
            print(dec.bf, dec.daughters_to_string())

    def getDecList(self, line):
        predeclist = self.decaylist[line].decays[:]
        declist = []
        for dec in predeclist:
            declist.append([[(line, dec)],dec])
##            print dec
##            declist.append([[],dec])
##         dlist = DaughterList()
##         dlist.add(line)
##         declist = [[[],Decay(1,dlist)]]
        decaytable = self.decaylist.copy()
        for j in self.termpart:
            del decaytable[j]
        last = []
        while last != declist:
            last = declist[:]
            recurseOneLevel(declist, decaytable)
        return declist


    def do_explain(self, line):
        part = input('Parent particle? ')
        if part not in self.decaylist:
            print('%s not known. Have you read the file yet?' % part)
            return
        final = input('Final state? ').split()
        finalhash = {}
        termpartcpy = self.termpart[:]
        for p in final:
            finalhash[p] = finalhash.get(p,0) + 1
            if p in self.decaylist and p not in self.termpart:
                self.termpart.append(p)
        declist = self.getDecList(part)
        self.termpart[:] = termpartcpy
        sublist = []
        for dec in declist:
##            print dec[1].daughters
            if dec[1].daughters == finalhash:
                sublist.append(dec)
        sublist.sort(key=lambda x: x[1].bf, reverse=True)
        for entry in sublist:
            print(entry[1].bf, end=' ')
            if len(entry[0]) == 0:
                print('Direct ')
            else:
                print('Chain: ')
                for e2 in entry[0]:
##                    print e2
                    print('\t', e2[0], '->', e2[1].daughters_to_string())


    def do_oneshot(self, line):
        declist = self.decaylist[line].decays[:]
        recurseOneLevel(declist, self.decaylist)
        for dec in declist:
            print(dec.bf, dec.daughters)

    def do_EOF(self, line):
        print()
        return True

def recurseOneLevel(decaylist, decaytable):
    """Do one sweep at replacing particles with daughters."""
    """decaylist should be tuple of [[(particle, decay) ...], finalstate]"""
    rv = []
    for decay in decaylist:
        # Only do one daughter!
        toexpand = None
        for dau in decay[1].daughters:
            if toexpand == None and dau in decaytable:
                toexpand = dau
        if toexpand == None:
            rv.append(decay)
        else:
            for subdec in decaytable[toexpand].decays:
                chain = decay[0][:]
                chain.append((toexpand, subdec))
                dlist = DaughterList()
                dlist.update(decay[1].daughters)
                dlist[toexpand] -= 1
                if dlist[toexpand] == 0:
                    del dlist[toexpand]
                for entry in subdec.daughters:
                    dlist[entry] = dlist.get(entry, 0) + subdec.daughters[entry]
                newdecay = Decay(decay[1].bf*subdec.bf, dlist)
                rv.append([chain, newdecay])
    decaylist[:] = rv

def compactDecayList(decaylist):
    rv = []; used_daughter_list = []
    for decay in decaylist:
        if decay[1].daughters not in used_daughter_list:
            rv.append(Decay(decay[1].bf, decay[1].daughters))
            used_daughter_list.append(decay[1].daughters)
        else:
            for dec2 in rv:
                if dec2.daughters == decay[1].daughters:
                    dec2.bf += decay[1].bf
    return rv


if __name__ == '__main__':
##     decparser().cmdloop(
##         """Hi! We are beginning the loop."""
##         )
##    q = decparser(stdin=file('/home/ponyisi/DECAY.DEC', 'r'), stdout=sys.stdout)
##    q.use_rawinput = False
##    q.cmdloop()
##    print q.decaylist
    interactive().cmdloop(
        "Hi! Welcome to the DECAY.DEC parser program.\n"
        "Blame ponyisi@lepp if any problems arise."
        )



# Copyright (c) 2018-2020, Eduardo Rodrigues and Henry Schreiner.
#
# Distributed under the 3-clause BSD license, see accompanying file LICENSE
# or https://github.com/scikit-hep/decaylanguage for details.

from __future__ import absolute_import

try:
    from functools import lru_cache
    cacher = lru_cache(maxsize=64)
except ImportError:
    from cachetools import cached, LFUCache
    cacher = cached(cache=LFUCache(maxsize=64))

from particle import Particle
from particle.converters import EvtGenName2PDGIDBiMap
from particle.converters import PDG2EvtGenNameMap,EvtGen2PDGNameMap
from particle.exceptions import MatchingIDNotFound


@cacher
def charge_conjugate_name(name, pdg_name=False):
    """
    Return the charge-conjugate particle name matching the given name.
    If no matching is found, return "ChargeConj(pname)".

    Note
    ----
    Search/match in order:
    1) Trivial case - does the name correspond to a self-conjugate particle?
       Only works for particles in the DB.
    2) Try to match the antiparticle looking for the opposite PDG ID
       in the list of PDG IDs - EvtGen names.
       This can deal with specific particles or badly-known particles
       not in the DB but not so rare in decay files.

    Parameters
    ----------
    name: str
        Input particle name.
    pdg_name: str, optional, default=False
        Input particle name is the PDG name,
        not the (default) EvtGen name.

    Returns
    -------
    out: str
        Either the EvtGen or PDG charge-conjugate particle name
        depending on the value of parameter `pdg_name`.
    """
    if pdg_name:
        try:
            ccname = charge_conjugate_name(PDG2EvtGenNameMap[name])
            # Convert the EvtGen name back to a PDG name, to match input type
            return EvtGen2PDGNameMap[ccname]
        except MatchingIDNotFound:  # Catch issue in PDG2EvtGenNameMap matching
            return 'ChargeConj({0})'.format(name)

    # Dealing only with EvtGen names at this stage
    try:
        return Particle.from_evtgen_name(name).invert().evtgen_name
    except:
        try:
            return EvtGenName2PDGIDBiMap[-EvtGenName2PDGIDBiMap[name]]
        except:
            return 'ChargeConj({0})'.format(name)



# Copyright (c) 2018-2020, Eduardo Rodrigues and Henry Schreiner.
#
# Distributed under the 3-clause BSD license, see accompanying file LICENSE
# or https://github.com/scikit-hep/decaylanguage for details.

from __future__ import absolute_import
from __future__ import print_function

from collections import Counter
from copy import deepcopy

from ..utils import charge_conjugate_name


class DaughtersDict(Counter):
    """
    Class holding a decay final state as a dictionary.
    It is a building block for the digital representation of full decay chains.

    Note
    ----
    This class assumes EvtGen particle names, though this assumption is only relevant
    for the `charge_conjugate` method.
    Otherwise, all other class methods smoothly deal with
    any kind of particle names (basically an iterable of strings).

    Example
    -------
    A final state such as 'K+ K- K- pi+ pi0' is stored as
    ``{'K+': 1, 'K-': 2, 'pi+': 1, 'pi0': 1}``.
    """

    def __init__(self, iterable=None, **kwds):
        """
        Default constructor.

        Note
        ----
        This class assumes EvtGen particle names, though this assumption is only relevant
        for the `charge_conjugate` method (refer to its documentation).
        Otherwise, all other class methods smoothly deal with
        any kind of particle names (basically an iterable of strings).

        Examples
        --------
        >>> # An empty final state
        >>> dd = DaughtersDict()

        >>> # Constructor from a dictionary
        >>> dd = DaughtersDict({'K+': 1, 'K-': 2, 'pi+': 1, 'pi0': 3})

        >>> # Constructor from a list of particle names
        >>> dd = DaughtersDict(['K+', 'K-', 'K-', 'pi+', 'pi0'])

        >>> # Constructor from a string representing the final state
        >>> dd = DaughtersDict('K+ K- pi0')
        """
        if iterable and isinstance(iterable, str):
            iterable = iterable.split()
        super(DaughtersDict, self).__init__(iterable, **kwds)

    def to_string(self):
        """
        Return the daughters as a string representation (ordered list of names).
        """
        return ' '.join(sorted(self.elements()))

    def to_list(self):
        """
        Return the daughters as an ordered list of names.
        """
        return sorted(list(self.elements()))

    def charge_conjugate(self, pdg_name=False):
        """
        Return the charge-conjugate final state.

        Parameters
        ----------
        pdg_name: str, optional, default=False
            Input particle name is the PDG name,
            not the (default) EvtGen name.

        Examples
        --------
        >>> dd = DaughtersDict({'K+': 2, 'pi0': 1})
        >>> dd.charge_conjugate()
        <DaughtersDict: ['K-', 'K-', 'pi0']>
        >>>
        >>> dd = DaughtersDict({'K_S0': 1, 'pi0': 1})
        >>> dd.charge_conjugate()
        <DaughtersDict: ['K_S0', 'pi0']>
        >>>
        >>> dd = DaughtersDict({'K(S)0': 1, 'pi0': 1})  # PDG names!
        >>> dd.charge_conjugate(pdg_name=True)
        <DaughtersDict: ['K(S)0', 'pi0']>
        """
        return self.__class__({charge_conjugate_name(p, pdg_name):n for p, n in self.items()})

    def __repr__(self):
        return "<{self.__class__.__name__}: {daughters}>".format(
                self=self, daughters=self.to_list())

    def __str__(self):
        return repr(self)

    def __len__(self):
        """
        Return the length, i.e. the number of final-state particles.

        Note
        ----
        This is generally *not* the number of dictionary elements.
        """
        return sum(n for n in self.values())

    def __add__(self, other):
        """
        Add two final states, particle-type-wise.
        """
        dd = super(DaughtersDict, self).__add__(other)
        return self.__class__(dd)

    def __iter__(self):
        return self.elements()


class DecayMode(object):
    """
    Class holding a particle decay mode, which is typically a branching fraction
    and a list of final-state particles (i.e. a list of DaughtersDict instances).
    The class can also contain metadata such as decay model and optional
    decay-model parameters, as defined for example in .dec decay files.

    This class is a building block for the digital representation
    of full decay chains.

    Note
    ----
    This class assumes EvtGen particle names, though this assumption is only
    relevant for the `charge_conjugate` method.
    Otherwise, all other class methods smoothly deal
    with any kind of particle names (basically an iterable of strings).
    """

    __slots__ = ("bf",
                 "daughters",
                 "metadata")

    def __init__(self, bf = 0, daughters = None, **info):
        """
        Default constructor.

        Parameters
        ----------
        bf: float, optional, default=0
            Decay mode branching fraction.
        daughters: iterable or DaughtersDict, optional, default=None
            The final-state particles.
        info: keyword arguments, optional
            Decay mode model information and/or user metadata (aka extra info)
            By default the following elements are always created:
            dict(model=None, model_params=None).
            The user can provide any metadata, see the examples below.

        Note
        ----
        This class assumes EvtGen particle names, though this assumption is only
        relevant for the `charge_conjugate` method.
        Otherwise, all other class methods smoothly deal
        with any kind of particle names (basically an iterable of strings).

        Examples
        --------
        >>> # A 'default' and hence empty, decay mode
        >>> dm = DecayMode()

        >>> # Decay mode with minimal input information
        >>> dd = DaughtersDict('K+ K-')
        >>> dm = DecayMode(0.5, dd)

        >>> # Decay mode with minimal input information, simpler version
        >>> dm = DecayMode(0.5, 'K+ K-')

        >>> # Decay mode with decay model information
        >>> dd = DaughtersDict('pi- pi0 nu_tau')
        >>> dm = DecayMode(0.2551, dd,
                           model='TAUHADNU',
                           model_params=[-0.108, 0.775, 0.149, 1.364, 0.400])

        >>> # Decay mode with user metadata
        >>> dd = DaughtersDict('K+ K-')
        >>> dm = DecayMode(0.5, dd, model='PHSP', study='toy', year=2019)
        """
        self.bf = bf
        self.daughters = DaughtersDict(daughters)

        self.metadata = dict(model='', model_params='')
        self.metadata.update(**info)

    @classmethod
    def from_dict(cls, decay_mode_dict):
        """
        Constructor from a dictionary of the form
        {'bf': <float>, 'fs': [...], ...}.
        These two keys are mandatory.

        Note
        ----
        This class assumes EvtGen particle names, though this assumption is only
        relevant for the `charge_conjugate` method.
        Otherwise, all other class methods smoothly deal
        with any kind of particle names (basically an iterable of strings).

        Examples
        --------
        >>> DecayMode.from_dict({'bf': 0.98823,
                                 'fs': ['gamma', 'gamma'],
                                 'model': 'PHSP',
                                 'model_params': ''})
        <DecayMode: daughters=gamma gamma, BF=0.98823>
        """
        dm = deepcopy(decay_mode_dict)

        try:
            bf = dm.pop('bf')
            daughters = dm.pop('fs')
        except:
            raise RuntimeError("Input not in the expected format!")

        return cls(bf=bf, daughters=daughters, **dm)

    @classmethod
    def from_pdgids(cls, bf, daughters, **info):
        """
        Constructor for a final state given as a list of particle PDG IDs.

        Note
        ----
        All particle names are internally saved as EvtGen names,
        to be consistent with the default class assumption, see class docstring.

        Examples
        --------
        >>> DecayMode.from_pdgids(0.5, [321, -321])
        <DecayMode: daughters=K+ K-, BF=0.5>
        >>>
        >>> DecayMode.from_pdgids(0.5, [310, 310])
        <DecayMode: daughters=K_S0 K_S0, BF=0.5>
        """
        # Check inputs
        try:
            from particle import PDGID, ParticleNotFound
            from particle.converters import EvtGenName2PDGIDBiMap
            daughters = [EvtGenName2PDGIDBiMap[PDGID(d)] for d in daughters]
        except ParticleNotFound:
            raise ParticleNotFound('Please check your input PDG IDs!')

        daughters = DaughtersDict(daughters)

        # Override the default settings with the user input, if any
        return cls(bf=bf, daughters=daughters, **info)

    def describe(self):
        """
        Make a nice high-density string for all decay-mode properties and info.
        """
        val = """Daughters: {daughters} , BF: {bf:<15.8g}
    Decay model: {model} {model_params}""".format(daughters=' '.join(self.daughters),
           bf=self.bf,
           model=self.metadata['model'],
           model_params=self.metadata['model_params']
                        if self.metadata['model_params'] is not None else '')

        keys = [k for k in self.metadata
              if k not in ('model', 'model_params')]
        if len(keys) > 0:
            val += "\n    Extra info:\n"
        for key in keys:
            val += "        {k}: {v}\n".format(k=key, v=self.metadata[key])

        return val

    def to_dict(self):
        """
        Return the decay mode as a dictionary in the format understood
        by the `DecayChainViewer` class.

        Examples
        --------
        >>> dm = DecayMode(0.5, 'K+ K- K- pi- pi0 nu_tau',
                           model='PHSP', study='toy', year=2019)
        >>> dm.to_dict()
        {'bf': 0.5,
         'fs': ['K+', 'K-', 'K-', 'nu_tau', 'pi-', 'pi0'],
         'model': 'PHSP',
         'model_params': '',
         'study': 'toy',
         'year': 2019}
        """
        d = {'bf': self.bf, 'fs': self.daughters.to_list()}
        d.update(self.metadata)
        if d['model_params'] is None:
            d['model_params'] = ''
        return d

    def charge_conjugate(self, pdg_name=False):
        """
        Return the charge-conjugate decay mode.

        Parameters
        ----------
        pdg_name: str, optional, default=False
            Input particle name is the PDG name,
            not the (default) EvtGen name.

        Examples
        --------
        >>> dm = DecayMode(1.0, 'K+ K+ pi-')
        >>> dm.charge_conjugate()
        <DecayMode: daughters=K- K- pi+, BF=1.0>
        >>>
        >>> dm = DecayMode(1.0, 'K_S0 pi+')
        >>> dm.charge_conjugate()
        <DecayMode: daughters=K_S0 pi-, BF=1.0>
        >>>
        >>> dm = DecayMode(1.0, 'K(S)0 pi+')  # PDG names!
        >>> dm.charge_conjugate(pdg_name=True)
        <DecayMode: daughters=K(S)0 pi-, BF=1.0>
        """
        return self.__class__(self.bf,
                              self.daughters.charge_conjugate(pdg_name),
                              **self.metadata)

    def __len__(self):
        """
        The "length" of a decay mode is the number of final-state particles.
        """
        return len(self.daughters)

    def __repr__(self):
        return "<{self.__class__.__name__}: daughters={daughters}, BF={bf}>".format(
                self=self,
                daughters=self.daughters.to_string() if len(self.daughters)>0 else '[]',
                bf=self.bf)

    def __str__(self):
        return repr(self)


class DecayChain(object):
    """
    Class holding a particle decay chain, which is typically a top-level decay
    (mother particle, branching fraction and final-state particles)
    and a set of sub-decays for any non-stable particle in the top-level decay.
    The whole chain can be seen as a mother particle and a list of decay modes.

    This class is the main building block for the digital representation
    of full decay chains.

    Note
    ----
    This class does not assume any kind of particle names (EvtGen, PDG).
    It is nevertheless advised to default use EvtGen names for consistency
    with the defaults used in the related classes `DecayMode` and `DaughtersDict`,
    unless there is a good motivation not to.
    """

    __slots__ = ("mother",
                 "decays")

    def __init__(self, mother, decays):
        """
        Default constructor.

        Parameters
        ----------
        mother: str, optional, default=None
            Input mother particle of the top-level decay.

        Examples
        --------
        >>> dm1 = DecayMode(0.0124, 'K_S0 pi0', model='PHSP')
        >>> dm2 = DecayMode(0.692, 'pi+ pi-')
        >>> dm3 = DecayMode(0.98823, 'gamma gamma')
        >>> dc = DecayChain('D0', {'D0':dm1, 'K_S0':dm2, 'pi0':dm3})
        """
        self.mother = mother
        self.decays = decays

    @classmethod
    def from_dict(cls, decay_chain_dict):
        """
        Constructor from a decay chain represented as a dictionary.
        The format is the same as that returned by
        `DecFileParser.build_decay_chains(...)`.
        """
        try:
            assert len(decay_chain_dict.keys()) == 1
        except:
            raise RuntimeError("Input not in the expected format!")

        has_no_subdecay = lambda ds: all([isinstance(p, str) for p in ds])

        def build_decay_modes(dc_dict):
            mother = list(dc_dict.keys())[0]
            dms = dc_dict[mother]

            for dm in dms:
                if has_no_subdecay(dm['fs']):
                    decay_modes[mother] = DecayMode.from_dict(dm)
                else:
                    d = deepcopy(dm)
                    for i in range(len(d['fs'])):
                        if isinstance(d['fs'][i], dict):
                            # Replace the element with the key and
                            # store the present decay mode ignoring sub-decays
                            d['fs'][i] = list(d['fs'][i].keys())[0]
                            # Recursively continue ...
                            build_decay_modes(dm['fs'][i])
                    # Create the decay mode now that none of its particles
                    # has a sub-decay
                    decay_modes[mother] = DecayMode.from_dict(d)


        decay_modes = dict()
        mother = list(decay_chain_dict.keys())[0]
        build_decay_modes(decay_chain_dict)

        return cls( mother, decay_modes)

    def top_level_decay(self):
        """
        Return the top-level decay as a `DecayMode` instance.
        """
        return self.decays[self.mother]

    @property
    def bf(self):
        """
        Branching fraction of the top-level decay.
        """
        return self.top_level_decay().bf

    @property
    def visible_bf(self):
        """
        Visible branching fraction of the whole decay chain.

        Note
        ----
        Calculation requires a flattening of the entire decay chain.
        """
        return self.flatten().bf

    @property
    def ndecays(self):
        """
        Return the number of decay modes including the top-level decay.
        """
        return len(self.decays)

    def print_as_tree(self):
        """
        Tree-structure like print of the entire decay chain.

        Examples
        --------
        >>> dm1 = DecayMode(0.028, 'K_S0 pi+ pi-')
        >>> dm2 = DecayMode(0.692, 'pi+ pi-')
        >>> dc = DecayChain('D0', {'D0':dm1, 'K_S0':dm2})
        >>> dc.print_as_tree()
        D0
        +--> K_S0
        |    +--> pi+
        |    +--> pi-
        +--> pi+
        +--> pi-

        >>> dm1 = DecayMode(0.0124, 'K_S0 pi0')
        >>> dm2 = DecayMode(0.692, 'pi+ pi-')
        >>> dm3 = DecayMode(0.98823, 'gamma gamma')
        >>> dc = DecayChain('D0', {'D0':dm1, 'K_S0':dm2, 'pi0':dm3})
        >>> dc.print_as_tree()
        D0
        +--> K_S0
        |    +--> pi+
        |    +--> pi-
        +--> pi0
             +--> gamma
             +--> gamma

        >>> dm1 = DecayMode(0.6770, 'D0 pi+')
        >>> dm2 = DecayMode(0.0124, 'K_S0 pi0')
        >>> dm3 = DecayMode(0.692, 'pi+ pi-')
        >>> dm4 = DecayMode(0.98823, 'gamma gamma')
        >>> dc = DecayChain('D*+', {'D*+':dm1, 'D0':dm2, 'K_S0':dm3, 'pi0':dm4})
        >>> dc.print_as_tree()
        D*+
        +--> D0
        |    +--> K_S0
        |    |    +--> pi+
        |    |    +--> pi-
        |    +--> pi0
        |         +--> gamma
        |         +--> gamma
        +--> pi+
        """
        indent = 4
        arrow = '+--> '
        bar = '|'

        # TODO: simplify logic and perform further checks
        def _print(decay_dict, depth=0, link=False, last=False):
            mother = list(decay_dict.keys())[0]
            prefix = bar if (link and depth>1) else ''
            prefix = prefix + ' '*indent*(depth-1)
            for i, i_decay in enumerate(decay_dict[mother]):
                print(prefix, arrow if depth > 0 else '', mother, sep='')
                fsps = i_decay['fs']
                n = len(list(fsps))
                depth += 1
                for j, fsp in enumerate(fsps):
                    prefix = bar if (link and depth>1) else ''
                    if last:
                        prefix = prefix + ' '*indent*(depth-1) + ' '
                    else:
                        prefix = (prefix+' '*indent)*(depth-1)
                    if isinstance(fsp, str):
                        print(prefix, arrow, fsp, sep='')
                    else:
                        _print(fsp, depth=depth, link=(link or (j<n-1)), last=(j==n-1))

        dc_dict = self.to_dict()
        _print(dc_dict)

    def to_dict(self):
        """
        Return the decay chain as a dictionary representation.
        The format is the same as `DecFileParser.build_decay_chains(...)`.

        Examples
        --------
        >>> dm1 = DecayMode(0.028, 'K_S0 pi+ pi-')
        >>> dm2 = DecayMode(0.692, 'pi+ pi-')
        >>> dc = DecayChain('D0', {'D0':dm1, 'K_S0':dm2})
        >>> dc.to_dict()
        {'D0': [{'bf': 0.028,
            'fs': [{'K_S0': [{'bf': 0.692,
                'fs': ['pi+', 'pi-'],
                'model': '',
                'model_params': ''}]},
             'pi+',
             'pi-'],
            'model': '',
            'model_params': ''}]}
        """
        def recursively_replace(mother):
            dm = self.decays[mother].to_dict()
            result = list()
            list_fsp = dm['fs']

            for pos, fsp in enumerate(list_fsp):
                if fsp in self.decays.keys():
                    list_fsp[pos] = recursively_replace(fsp)
                else:
                    pass

            result.append(dm)
            d = {mother:result}
            return d

        return recursively_replace(self.mother)

    def flatten(self, stable_particles=[]):
        """
        Flatten the decay chain replacing all intermediate, decaying particles,
        with their final states.

        Parameters
        ----------
        stable_particles: iterable, optional, default=[]
            If provided, ignores the sub-decays of the listed particles,
            considering them as stable.

        Examples
        --------
        >>> dm1 = DecayMode(0.0124, 'K_S0 pi0', model='PHSP')
        >>> dm2 = DecayMode(0.692, 'pi+ pi-')
        >>> dm3 = DecayMode(0.98823, 'gamma gamma')
        >>> dc = DecayChain('D0', {'D0':dm1, 'K_S0':dm2, 'pi0':dm3})
        >>>
        >>> dc.flatten()
        <DecayChain: D0 -> gamma gamma pi+ pi- (0 sub-decays), BF=0.008479803984>
        >>> dc.flatten().to_dict()
        {'D0': [{'bf': 0.008479803984,
           'fs': ['gamma', 'gamma', 'pi+', 'pi-'],
           'model': 'PHSP',
           'model_params': ''}]}

        >>> dc.flatten(stable_particles=['K_S0', 'pi0']).decays
        {'D0': <DecayMode: daughters=K_S0 pi0, BF=0.0124>}
        """
        vis_bf = 1.
        fs = DaughtersDict()
        keys = self.decays.keys()
        for k in keys:
            down_one_level = True
            if k in stable_particles:
                continue
            while down_one_level:
                vis_bf *= self.decays[k].bf
                fs += self.decays[k].daughters
                fs[k] -= 1
                down_one_level = k in fs.elements()

        return DecayChain(self.mother,
            {self.mother:DecayMode(vis_bf,
                                   fs,
                                   **self.top_level_decay().metadata)
                                   }
            )

    def __repr__(self):
        if self.mother is None:
            return "Decay mode: undefined"

        return "<{self.__class__.__name__}: {mother} -> {tldecay} ({n} sub-decays), BF={bf}>".format(
                self=self,
                mother=self.mother,
                tldecay=self.top_level_decay().daughters.to_string(),
                n=len(self.decays)-1,
                bf=self.bf)

    def __str__(self):
        return repr(self)



# Copyright (c) 2020, Eduardo Rodrigues and Henry Schreiner.
#
# Distributed under the 3-clause BSD license, see accompanying file LICENSE
# or https://github.com/scikit-hep/decaylanguage for details.

import pytest
from lark import UnexpectedToken

try:
    from pathlib2 import Path
except ImportError:
    from pathlib import Path

from decaylanguage.dec.dec import DecFileParser


DIR = Path(__file__).parent.resolve()


def test_issue_90():
    with pytest.raises(UnexpectedToken):
        p = DecFileParser(DIR / '../data/test_issue90.dec')
        p.parse()



# Copyright (c) 2018-2020, Eduardo Rodrigues and Henry Schreiner.
#
# Distributed under the 3-clause BSD license, see accompanying file LICENSE
# or https://github.com/scikit-hep/decaylanguage for details.

'''
This is a GooFit adaptor for amplitude chain.
'''

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from enum import Enum

import pandas as pd

from particle import SpinType
from particle.particle.utilities import programmatic_name

from ..utils import LineFailure
from .amplitudechain import LS
from .amplitudechain import AmplitudeChain


class SF_4Body(Enum):
    DtoPP1_PtoSP2_StoP3P4 = 0
    DtoPP1_PtoVP2_VtoP3P4 = 1
    DtoV1V2_V1toP1P2_V2toP3P4_S = 2
    DtoV1V2_V1toP1P2_V2toP3P4_P = 3
    DtoV1V2_V1toP1P2_V2toP3P4_D = 4
    DtoAP1_AtoVP2_VtoP3P4 = 5
    DtoAP1_AtoVP2Dwave_VtoP3P4 = 6
    DtoVS_VtoP1P2_StoP3P4 = 7
    DtoV1P1_V1toV2P2_V2toP3P4 = 8
    DtoAP1_AtoSP2_StoP3P4 = 9
    DtoTP1_TtoVP2_VtoP3P4 = 10
    FF_12_34_L1 = 11
    FF_12_34_L2 = 12
    FF_123_4_L1 = 13
    FF_123_4_L2 = 14
    ONE = 15


known_spinfactors = {
    'DtoA1P1_A1toS2P2_S2toP3P4':
    (SF_4Body.DtoAP1_AtoSP2_StoP3P4,),
    'DtoA1P1_A1toV2P2Dwave_V2toP3P4':
    (SF_4Body.DtoAP1_AtoVP2Dwave_VtoP3P4,),
    'DtoA1P1_A1toV2P2_V2toP3P4':
    (SF_4Body.DtoAP1_AtoVP2Dwave_VtoP3P4,),
    'DtoS1S2_S1toP1P2_S2toP3P4':
    (SF_4Body.ONE,),
    'DtoT1P1_T1toV2P2_V2toP3P4':
    (SF_4Body.DtoTP1_TtoVP2_VtoP3P4,),
    'DtoV1S2_V1toP1P2_S2toP3P4':
    (SF_4Body.DtoVS_VtoP1P2_StoP3P4,),
    'DtoV1V2_V1toP1P2_V2toP3P4':
    (SF_4Body.DtoV1V2_V1toP1P2_V2toP3P4_S,),
    'DtoV1V2_V1toP1P2_V2toP3P4_D':
    (SF_4Body.DtoV1V2_V1toP1P2_V2toP3P4_D,),
    'DtoV1V2_V1toP1P2_V2toP3P4_P':
    (SF_4Body.DtoV1V2_V1toP1P2_V2toP3P4_P,),
    'Dtos1P1_s1toS2P2_S2toP3P4':
    (SF_4Body.DtoPP1_PtoSP2_StoP3P4,),
    'Dtos1P1_s1toV2P2_V2toP3P4':
    (SF_4Body.DtoPP1_PtoVP2_VtoP3P4,),
}


def sprint(stype):
    if stype in {SpinType.PseudoTensor, SpinType.PseudoScalar}:
        return stype.name[6].lower()
    else:
        return stype.name[0]


class DecayStructure(Enum):
    FF_12_34 = 0
    FF_1_2_34 = 1


class GooFitChain(AmplitudeChain):
    __slots__ = ()

    # pars: pd.DataFrame
    # consts: pd.DataFrame
    pars = None
    consts = None

    @classmethod
    def make_intro(cls, all_states):

        header = '    // Event type: {} ->  '.format(all_states[0])
        header += '   '.join('{} ({})'.format(b, a) for a, b in enumerate(all_states[1:]))

        header += '''\n
    std::vector<std::vector<Lineshape*>> line_factor_list;
    std::vector<std::vector<SpinFactor*>> spin_factor_list;
    std::vector<Amplitude*> amplitudes_list;

'''

        final_particles = set(all_states)

        for particle in final_particles:
            name = particle.programmatic_name.upper()
            header += '    constexpr fptype {name:8} {{ {particle.mass:<14.8g} }};\n'.format(
                name=name, particle=particle)

        header += '\n'

        for particle in cls.all_particles - final_particles:
            name = particle.programmatic_name
            header += '    Variable {name:15} {{ {nameQ:21}, {particle.mass:<10.8g} }};\n'.format(
                name=name+'_M', nameQ='"'+name+'_M"', particle=particle)
            header += '    Variable {name:15} {{ {nameQ:21}, {particle.width:<10.8g} }};\n'.format(
                name=name+'_W', nameQ='"'+name+'_W"', particle=particle)

        header += '\n'
        header += '    DK3P_DI.meson_radius = 5;\n'
        header += '    DK3P_DI.particle_masses = {{{}}};\n'.format(
            ', '.join(x.programmatic_name.upper() for x in all_states))

        return header

    @property
    def decay_structure(self):
        if len(self[0]) == 2 and len(self[1]) == 2:
            return DecayStructure.FF_12_34
        else:
            return DecayStructure.FF_1_2_34

    @property
    def formfactor(self):
        norm = self.decay_structure == DecayStructure.FF_12_34
        if self.L == 0:
            return None
        elif self.L == 1:
            return SF_4Body.FF_12_34_L1 if norm else SF_4Body.FF_123_4_L1
        elif self.L == 2:
            return SF_4Body.FF_12_34_L2 if norm else SF_4Body.FF_123_4_L2
        else:
            raise NotImplementedError("L = {self.L} is not implemented".format(self=self))

    def spindetails(self):
        if self.decay_structure == DecayStructure.FF_12_34:
            a = "{0}1".format(sprint(self[0].particle.spin_type))
            b = "{0}2".format(sprint(self[1].particle.spin_type))
            return ("Dto{a}{b}_{a}toP1P2_{b}toP3P4"
                    + ("_{self.spinfactor}" if self.spinfactor and self.spinfactor != 'S' else "")
                    ).format(self=self, a=a, b=b)
        else:
            a = "{0}1".format(sprint(self[0].particle.spin_type))
            if self[0].daughters:
                b = "{0}2".format(sprint(self[0][0].particle.spin_type))
            else:
                raise LineFailure(self, "{0} has no daughters".format(self[0]))
            wave = "{self[0].spinfactor}wave".format(
                self=self) if self[0].spinfactor and self[0].spinfactor != 'S' else ""
            return "Dto{a}P1_{a}to{b}P2{wave}_{b}toP3P4".format(a=a, b=b, wave=wave)

    @property
    def spinfactors(self):
        if self.spindetails() in known_spinfactors:
            spinfactor = list(known_spinfactors[self.spindetails()])
            if self.L > 0:
                spinfactor.append(self.formfactor)
            return spinfactor

        raise LineFailure(self, "Spinfactors not currenly included!: {spindet}".format(
            spindet=self.spindetails()))

        # if self.decay_structure == DecayStructure.FF_12_34 :
        #    if (self[0].particle.spin_type in {SpinType.Vector, SpinType.Axial}
        #        and self[1].particle.spin_type in {SpinType.Vector, SpinType.Axial}):
        #
        #        if self.spinfactor == 'D':
        #            return (SF_4Body.DtoV1V2_V1toP1P2_V2toP3P4_D, SF_4Body.FF_12_34_L2)
        #        elif self.spinfactor == 'P':
        #            return (SF_4Body.DtoV1V2_V1toP1P2_V2toP3P4_P, SF_4Body.FF_12_34_L1)
        #        elif self.spinfactor == 'S':
        #            return (SF_4Body.DtoV1V2_V1toP1P2_V2toP3P4_S,)
        # else:
        #    if (self[0].particle.spin_type == SpinType.Axial and
        #        self[0][0].particle.spin_type == SpinType.Vector):
        #        if self.spinfactor == 'D':
        #            return (SF_4Body.DtoAP1_AtoVP2Dwave_VtoP3P4, SF_4Body.FF_12_34_L2)
        #        else:
        #            return (SF_4Body.DtoAP1_AtoVP2_VtoP3P4, SF_4Body.FF_12_34_L1) # L1?

    @classmethod
    def make_pars(cls):
        headerlist = []
        header = ''

        for name, par in cls.pars.iterrows():
            pname = programmatic_name(name)
            if par.fix == 2:
                headerlist.append('    Variable {pname} {{"{name}", {par.value}, {par.error} }};'.format(
                    pname=pname, name=name, par=par))
            else:
                headerlist.append('    Variable {pname} {{"{name}", {par.value} }};'.format(
                    pname=pname, name=name, par=par))

        def strip_pararray(pars, begin, convert=lambda x: x):
            mysplines = pars.index[pars.index.str.contains(begin, regex=False)]
            vals = convert(mysplines.str.slice(len(begin))).astype(int)
            series = pd.Series(mysplines, vals).sort_index()
            return ',\n'.join(series.map(lambda x: '        '+programmatic_name(x)))

        if not GooFitChain.consts.empty:
            splines = GooFitChain.consts.index[GooFitChain.consts.index.str.contains("Spline")]
            splines = set(splines.str.rstrip("::Spline::N").str.rstrip(
                "::Spline::Min").str.rstrip("::Spline::Max"))

            for spline in splines:
                header += '\n    std::vector<Variable> ' + programmatic_name(spline) + "_SplineArr {{\n"
                header += strip_pararray(GooFitChain.pars,
                                         "{spline}::Spline::Gamma::".format(spline=spline))
                header += '\n    }};\n'

        f_scatt = GooFitChain.pars.index[GooFitChain.pars.index.str.contains("f_scatt")]
        if len(f_scatt):
            header += '\n    std::array<Variable, 5> f_scatt {{\n'
            header += strip_pararray(GooFitChain.pars, "f_scatt")
            header += '\n    }};\n'

        IS_mat = GooFitChain.pars.index[GooFitChain.pars.index.str.contains("IS_p")]
        if len(IS_mat):
            names = ("pipi", "KK", "4pi", "EtaEta", "EtapEta", "mass")

            def convert(x):
                i = x.str.split('_').str[0]
                j = x.str.split('_').str[1].map(lambda x: names.index(x))
                return i.astype(int)*6 + j.astype(int)
            header += '\n    std::array<Variable, 5*6> IS_poles {{\n'
            header += strip_pararray(GooFitChain.pars, "IS_p", convert)
            header += '\n    }};\n'

        return '\n'.join(headerlist) + '\n' + header

    def make_lineshape(self, structure):
        name = self.name
        par = self.particle.programmatic_name
        a = structure[0]+1
        b = structure[1]+1
        L = self.L
        radius = 5.0 if 'c' in self.particle.quarks.lower() else 1.5

        if self.ls_enum == LS.RBW:
            return 'new Lineshapes::RBW("{name}", {par}_M, {par}_W, {L}, M_{a}{b}, FF::BL2)'.format(name=name, par=par, L=L, a=a, b=b)
        elif self.ls_enum == LS.GSpline:
            min = self.__class__.consts.loc["{self.name}::Spline::Min".format(self=self), "value"]
            max = self.__class__.consts.loc["{self.name}::Spline::Max".format(self=self), "value"]
            N = self.__class__.consts.loc["{self.name}::Spline::N".format(self=self), "value"]
            AdditionalVars = programmatic_name(self.name) + "_SplineArr"
            return '''new Lineshapes::GSpline("{name}", {par}_M, {par}_W, {L}, M_{a}{b}, FF::BL2,
            {radius}, {AdditionalVars}, Lineshapes::spline_t({min},{max},{N}))'''.format(
                    name=name, par=par, L=L, a=a, b=b, radius=radius, AdditionalVars=AdditionalVars, min=min, max=max, N=int(N))

        elif self.ls_enum == LS.kMatrix:
            _, poleprod, pterm = self.lineshape.split('.')
            is_pole = 'true' if poleprod == 'pole' else 'false'
            return '''new Lineshapes::kMatrix("{name}", {pterm}, {is_pole},
            sA0, sA, s0_prod, s0_scatt,
            f_scatt, IS_poles,
            {par}_M, {par}_W, {L}, M_{a}{b}, FF::BL2, {radius})'''.format(
                    name=name, pterm=pterm, is_pole=is_pole, par=par, L=L, a=a, b=b, radius=radius)

        elif self.ls_enum == LS.FOCUS:
            _, mod = self.lineshape.split('.')
            return ('new Lineshapes::FOCUS("{name}", Lineshapes::FOCUS::Mod::{mod},'
                    ' {par}_M, {par}_W, {L}, M_{a}{b}, FF::BL2, {radius})').format(
                name=name, mod=mod, par=par, L=L, a=a, b=b, radius=radius)

        else:
            raise NotImplementedError(
                "Unimplemented GooFit Lineshape {self.ls_enum.name}".format(self=self))

    def make_spinfactor(self, final_states):
        spin_factors = self.spinfactors

        intro = '    spin_factor_list.push_back(std::vector<SpinFactor*>({\n'
        factor = []
        for structure in self.list_structure(final_states):
            if not spin_factors:
                factor.append(
                    '        // TODO: Spin factor not implemented yet for {spindet}'.format(spindet=self.spindetails()))
            else:
                for spin_factor in spin_factors:
                    structure_list = ', '.join(map(str, structure))
                    factor.append('        new SpinFactor("SF", SF_4Body::{spin_factor.name:37}, {structure_list})'.format(
                        spin_factor=spin_factor, structure_list=structure_list))
        exit = '\n    }));\n'
        return intro + ',\n'.join(factor) + exit

    def make_linefactor(self, final_states):
        intro = '    line_factor_list.push_back(std::vector<Lineshape*>{\n'
        factor = []
        for structure in self.list_structure(final_states):
            for sub in self.vertexes:
                factor.append('        ' + sub.make_lineshape(structure))
        exit = '\n    });\n'
        return intro + ',\n'.join(factor) + exit

    def make_amplitude(self, final_states):
        n = len(self.list_structure(final_states))
        fix = 'true' if self.fix else 'false'
        return ('    amplitudes_list.push_back(new Amplitude{{\n'
                '        "{self!s}",\n'
                '        mkvar("{self!s}_r", {fix}, {self.amp.real:.6}, {self.err.real:.6}),\n'
                '        mkvar("{self!s}_i", {fix}, {self.amp.imag:.6}, {self.err.imag:.6}),\n'
                '        line_factor_list.back(),\n'
                '        spin_factor_list.back(),\n'
                '        {n}}});\n\n'
                '    DK3P_DI.amplitudes_B.push_back(amplitudes_list.back());'.format(
                    self=self, fix=fix, n=n))

    def to_goofit(self, final_states):
        return ('    // ' + str(self) + '\n\n'
                + self.make_spinfactor(final_states) + '\n'
                + self.make_linefactor(final_states) + '\n'
                + self.make_amplitude(final_states))

    @classmethod
    def read_ampgen(cls, *args, **kargs):
        line_arr, GooFitChain.pars, GooFitChain.consts, all_states = super(
            GooFitChain, cls).read_ampgen(*args, **kargs)
        return line_arr, all_states



# Copyright (c) 2018-2020, Eduardo Rodrigues and Henry Schreiner.
#
# Distributed under the 3-clause BSD license, see accompanying file LICENSE
# or https://github.com/scikit-hep/decaylanguage for details.


# Convenient access to the version number
from ._version import __version__

# Direct access to decay file parsing tools
from .dec import DecFileParser

# Direct access to decay chain visualization tools
from .decay import DecayChainViewer

# Direct access to decay chain representation classes
from .decay import DaughtersDict, DecayMode, DecayChain



# Copyright (c) 2018-2020, Eduardo Rodrigues and Henry Schreiner.
#
# Distributed under the 3-clause BSD license, see accompanying file LICENSE
# or https://github.com/scikit-hep/decaylanguage for details.

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function


def iter_flatten(iterable):
    '''
    Flatten nested tuples and lists
    '''
    for e in iterable:
        if isinstance(e, (list, tuple)):
            for f in iter_flatten(e):
                yield f
        else:
            yield e


def split(x):
    '''
    Break up a comma separated list, but respect curly brackets.

    For example:
    this, that { that { this, that } }
    would only break on the first comma, since the second is in a {} list
    '''

    c = 0
    i = 0
    out = []
    while len(x) > 0:
        if i + 1 == len(x):
            out.append(x[:i + 1])
            return out
        elif (x[i] == ',' and c == 0):
            out.append(x[:i])
            x = x[i + 1:]
            i = -1
        elif x[i] == '{':
            c += 1
        elif x[i] == '}':
            c -= 1
        i += 1


def filter_lines(matcher, input):
    '''
    Filter out lines into new variable if they match a regular expression
    '''
    output = [matcher.match(l).groupdict() for l in input if matcher.match(l) is not None]
    input = [l for l in input if matcher.match(l) is None]
    return output, input



# Copyright (c) 2018-2020, Eduardo Rodrigues and Henry Schreiner.
#
# Distributed under the 3-clause BSD license, see accompanying file LICENSE
# or https://github.com/scikit-hep/decaylanguage for details.

from particle import Particle

from decaylanguage.modeling.goofit import GooFitChain


def test_simple():
    lines, all_states = GooFitChain.read_ampgen(text='''

    # This is a test (should not affect output)

    EventType D0 K- pi+ pi+ pi-

    D0[D]{K*(892)bar0{K-,pi+},rho(770)0{pi+,pi-}} 2 1         0          2 0         0
    ''')

    assert Particle.from_pdgid(421) == all_states[0]  # D0
    assert Particle.from_pdgid(-321) == all_states[1]  # K-
    assert Particle.from_pdgid(211) == all_states[2]  # pi+
    assert Particle.from_pdgid(211) == all_states[3]  # pi+
    assert Particle.from_pdgid(-211) == all_states[4]  # pi-

    assert len(lines) == 1
    line, = lines



#!/usr/bin/env python
# coding: utf-8
# Copyright (c) 2018-2020, Eduardo Rodrigues and Henry Schreiner.
#
# Distributed under the 3-clause BSD license, see accompanying file LICENSE
# or https://github.com/scikit-hep/decaylanguage for details.

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from plumbum import cli

from decaylanguage.modeling.ampgen2goofit import ampgen2goofit


class DecayLanguageDecay(cli.Application):
    generator = cli.SwitchAttr(['-G', '--generator'], cli.Set('goofit'), mandatory=True)

    def main(self, filename):
        if self.generator == 'goofit':
            ampgen2goofit(filename)


def main():
    DecayLanguageDecay.run()


if __name__ == "__main__":
    main()



# Copyright (c) 2018-2020, Eduardo Rodrigues and Henry Schreiner.
#
# Distributed under the 3-clause BSD license, see accompanying file LICENSE
# or https://github.com/scikit-hep/decaylanguage for details.

from decaylanguage.modeling.ampgen2goofit import ampgen2goofit

try:
    from pathlib2 import Path
except ImportError:
    from pathlib import Path

import pytest


DIR = Path(__file__).parent.resolve()


@pytest.mark.skip
def test_full_convert():
    text = ampgen2goofit(DIR / '../models/DtoKpipipi_v2.txt', ret_output=True)
    with (DIR / 'output/DtoKpipipi_v2.cu').open() as f:
        assert (set(x.strip() for x in text.splitlines() if 'Generated on' not in x)
                == set(x.strip() for x in f.readlines() if 'Generated on' not in x))
